# Copyright (c) 2025, NVIDIA CORPORATION.
# SPDX-License-Identifier: BSD-3-Clause

import inspect
from typing import (
    Any,
    Callable,
    Dict,
    Generator,
    Generic,
    Hashable,
    Iterator,
    List,
    Optional,
    Tuple,
    TypeVar,
    Union,
)

from megatron.energon.edataclass import edataclass
from megatron.energon.errors import SYSTEM_EXCEPTIONS, FatalSampleError
from megatron.energon.flavors.base_dataset import (
    FlexState,
    SavableDataset,
    set_sample_restore_key,
)
from megatron.energon.savable import Savable
from megatron.energon.source_info import SourceInfo
from megatron.energon.worker import WorkerConfig
from megatron.energon.wrappers._log_exception import log_exception
from megatron.energon.wrappers.base import BaseWrapperDataset, SampleIndex
from megatron.energon.wrappers.buffer import SavableSampleBuffer
from megatron.energon.wrappers.skip import SkipSample

T_batch = TypeVar("T_batch", covariant=True)
T_batch_sample = TypeVar("T_batch_sample", covariant=True)


@edataclass
class Bucket(Savable, Generic[T_batch_sample]):
    batch_size: int

    samples: SavableSampleBuffer[T_batch_sample]

    def save_state(self) -> FlexState:
        return FlexState(
            batch_size=self.batch_size,
            samples=self.samples.save_state(),
        )

    def restore_state(self, state: FlexState):
        self.batch_size = state["batch_size"]
        self.samples.restore_state(state["samples"])


class GroupBatchDataset(
    BaseWrapperDataset[T_batch_sample, T_batch], Generic[T_batch_sample, T_batch]
):
    """This dataset wrapper transforms a dataset of samples into a dataset of batches, grouped by some criterion.
    The length is not correct, as this function can not predict the number of batches as there is no fixed batch size,
    instead it returns the inner dataset size.
    An example use case is: Image-Text samples, which are to be grouped by the image size into three
    size categories (e.g. 128x128, 256x256, 512x512) for efficient augmentation and batching.
    """

    dataset: SavableDataset[T_batch_sample]
    sample_group_key: Callable[[T_batch_sample], Tuple[Hashable, Optional[int]]]
    batcher: Callable[[List[T_batch_sample]], T_batch]
    drop_last: bool
    error_handler: Callable[[Exception, List[T_batch_sample], list[SourceInfo]], None]
    _group_key_sample_index: SampleIndex
    _batch_sample_index: SampleIndex
    _buckets: Dict[Hashable, Bucket[T_batch_sample]]

    def __init__(
        self,
        dataset: SavableDataset[T_batch_sample],
        fixed_batch_size: Optional[int],
        sample_group_key: Callable[[T_batch_sample], Tuple[Hashable, Optional[int]]],
        batcher: Callable[[List[T_batch_sample]], T_batch],
        *,
        batcher_stateless: bool = False,
        batcher_config: Optional[Union[Dict[str, Any], Callable[[], Dict[str, Any]]]] = None,
        drop_last: bool = False,
        error_handler: Callable[
            [Exception, List[T_batch_sample], list[SourceInfo]], None
        ] = log_exception,
        failure_tolerance: Optional[int] = 100,
        worker_config: WorkerConfig,
    ):
        """Construct a GroupBatchDataset.

        Args:
            dataset: The input dataset to wrap
            fixed_batch_size: Fixed batch size to use for all buckets. If None, the batch size is determined by the sample_group_key function.
            sample_group_key: Function which determines the bucket of a sample.
            batcher: Function which combines separate samples into a single object. May raise
                :exc:`megatron.energon.SkipSample` to skip a sample.
            drop_last: If True, the last batch is dropped if it is smaller than the batch size.
            error_handler: Handler for errors. Defaults to logging and ignoring the exception.
            failure_tolerance: The number of consecutive failures after which the dataset is considered broken.
            worker_config: Configuration for the workers.
        """
        super().__init__(dataset, worker_config=worker_config)
        self.fixed_batch_size = fixed_batch_size
        self.sample_group_key = sample_group_key
        self.batcher = batcher
        self.batcher_stateless = batcher_stateless
        self.batcher_config = batcher_config
        self.drop_last = drop_last
        self.error_handler = error_handler
        self.failure_tolerance = failure_tolerance

        self.reset_state_own()

        assert not inspect.isgeneratorfunction(batcher), (
            f"Batcher {batcher} must not be a generator function for grouped batching."
        )

    def reset_state_own(self) -> None:
        self._group_key_sample_index = SampleIndex(self.worker_config, src=self)
        self._batch_sample_index = SampleIndex(self.worker_config, src=self)
        self._buckets = {}

    def __len__(self):
        # Return an upper bound. This is for sure not correct.
        return len(self.dataset)

    def __iter__(self) -> Iterator[T_batch]:
        buckets = self._buckets

        last_batch_failures = 0

        if buckets is None:
            buckets = self._buckets = dict()

        # Load saved state if available
        for bucket in buckets.values():
            bucket.samples.worker_start()

        # print(f"[wrk={worker_idx}, s={self._batch_sample_index.current_idx}] initial GroupBatchDataset state:\n", end="")
        # for bucket_key, bucket in buckets.items():
        #     print(f"[wrk={worker_idx}, s={self._batch_sample_index.current_idx}] - Bucket [{bucket_key}] (bs={bucket.batch_size}, len(samples)={len(bucket.samples)}):\n", end="")
        #     bucket.samples.debug_print("    ")
        # print(f"[wrk={worker_idx}, s={self._batch_sample_index.current_idx}] initial done\n", end="")

        def flush(bucket: Bucket[T_batch_sample]) -> Generator[T_batch, None, None]:
            nonlocal last_batch_failures
            # Debug print the state
            # print(f"[wrk={worker_idx}, s={self._batch_sample_index.current_idx}] flush GroupBatchDataset state:\n", end="")
            # for dbg_bucket_key, dbg_bucket in buckets.items():
            #     print(f"[wrk={worker_idx}, s={self._batch_sample_index.current_idx}] - Bucket [{dbg_bucket_key}{'*' if dbg_bucket_key == bucket_key else ''}] (bs={dbg_bucket.batch_size}, len(samples)={len(dbg_bucket.samples)}):\n", end="")
            #     dbg_bucket.samples.debug_print("    ")
            batch_items, sample_restore_keys = bucket.samples.flush()
            # print(f"[wrk={worker_idx}, s={self._batch_sample_index.current_idx}] flushed: len(batch)={len(batch_items)} len(samples)={len(bucket.samples)}\n", end="")
            try:
                with self._batch_sample_index.ctx() as sample_idx:
                    batch_sample = self.batcher(batch_items)
                    assert not isinstance(batch_sample, Generator), (
                        f"Batcher {self.batcher} returned a generator, which is not supported for grouped batching yet."
                    )
                last_batch_failures = 0
                set_sample_restore_key(batch_sample, sample_idx, *sample_restore_keys, src=self)
                yield batch_sample
            except SkipSample:
                pass
            except SYSTEM_EXCEPTIONS:
                raise FatalSampleError.from_sample(batch_items)
            except Exception as e:
                self.error_handler(e, batch_items)
                last_batch_failures += 1
                if (
                    self.failure_tolerance is not None
                    and last_batch_failures >= self.failure_tolerance
                ):
                    raise FatalSampleError.from_sample(
                        batch_items,
                        f"GroupBatchDataset {self.batcher} failed {last_batch_failures} times in a row. Likely your code or dataset are broken.",
                    )

        # Add samples to the buckets
        for sample in self.dataset:
            try:
                with self._group_key_sample_index.ctx():
                    bucket_key, batch_size = self.sample_group_key(sample)
                    assert (batch_size is None) != (self.fixed_batch_size is None), (
                        f"A sample in group for key {bucket_key} returned batch size {batch_size}, but fixed "
                        f"batch size is set to {self.fixed_batch_size}. One of the two should be None."
                    )
                    if self.fixed_batch_size is not None:
                        batch_size = self.fixed_batch_size
            except SkipSample:
                continue
            except SYSTEM_EXCEPTIONS:
                raise FatalSampleError.from_sample(sample)
            except Exception as e:
                self.error_handler(e, [sample])
                continue
            bucket = buckets.get(bucket_key)
            if bucket is None:
                assert batch_size is not None
                buckets[bucket_key] = bucket = Bucket(
                    batch_size=batch_size,
                    samples=SavableSampleBuffer(self.dataset, worker_config=self.worker_config),
                )
            else:
                assert bucket.batch_size == batch_size, (
                    f"Got different batch size for group {bucket_key}: {bucket.batch_size} != {batch_size}."
                )
            bucket.samples.append(sample)
            if len(bucket.samples) >= bucket.batch_size:
                yield from flush(bucket)
        # Flush out last samples
        if not self.drop_last:
            for bucket in buckets.values():
                if len(bucket.samples) > 0:
                    yield from flush(bucket)
        # Clear the buckets
        self._buckets.clear()

    def save_state(self) -> FlexState:
        return FlexState(
            bucket_sample_index=self._group_key_sample_index.save_state(),
            batch_sample_index=self._batch_sample_index.save_state(),
            buckets={key: bucket.save_state() for key, bucket in self._buckets.items()},
            **super().save_state(),
        )

    def restore_state(self, state: FlexState) -> None:
        super().restore_state(state)

        self._group_key_sample_index.restore_state(state["bucket_sample_index"])
        self._batch_sample_index.restore_state(state["batch_sample_index"])
        for key, bucket_state in state["buckets"].items():
            self._buckets[key] = Bucket(
                batch_size=-1,
                samples=SavableSampleBuffer(self.dataset, worker_config=self.worker_config),
            )
            self._buckets[key].restore_state(bucket_state)

    def can_restore_sample(self) -> bool:
        return super().can_restore_sample() and self.batcher_stateless

    def assert_can_restore(self) -> None:
        assert self.batcher_stateless, (
            f"Batcher {self.batcher} must be stateless to restore samples"
        )
        super().assert_can_restore()

    def restore_sample(self, index: Tuple[Union[str, int, tuple], ...]) -> T_batch:
        self.assert_can_restore()
        id, sample_idx, *sample_restore_keys = index
        assert id == type(self).__name__
        batch = [self.dataset.restore_sample(inner_idx) for inner_idx in sample_restore_keys]
        with self._batch_sample_index.ctx(sample_idx):
            batch_sample = self.batcher(batch)
        set_sample_restore_key(batch_sample, sample_idx, *sample_restore_keys, src=self)
        return batch_sample

    def config(self) -> Dict[str, Any]:
        return {
            "type": type(self).__qualname__,
            "bucket": self._function_config(self.sample_group_key),
            "batcher": self._function_config(self.batcher),
            **(
                {
                    "batcher_config": (
                        self.batcher_config()
                        if callable(self.batcher_config)
                        else self.batcher_config
                    )
                }
                if self.batcher_config
                else {}
            ),
            "batcher_stateless": self.batcher_stateless,
            "drop_last": self.drop_last,
            "error_handler": self._function_config(self.error_handler),
            "worker_config": self.worker_config.config(),
            "dataset": self.dataset.config(),
        }

    def __str__(self):
        return f"GroupBatchDataset(bucket={self.sample_group_key}, batcher={self.batcher}, drop_last={self.drop_last}, dataset={self.dataset})"



# Copyright (c) 2025, NVIDIA CORPORATION.
# SPDX-License-Identifier: BSD-3-Clause

import gc
from typing import Any, Dict, Generic, Iterator, TypeVar

import torch
import torch.utils.data
import torch.utils.data.dataloader
from torch.distributed._shard.sharded_tensor import ShardedTensorBase
from torch.distributed.distributed_c10d import reduce_op

from megatron.energon.flavors.base_dataset import SavableDataset
from megatron.energon.worker import WorkerConfig
from megatron.energon.wrappers.base import BaseWrapperDataset

T_sample = TypeVar("T_sample")

_frozen_cuda_tensors = set()
_frozen_cuda_tensors_initialized = False


GC_DEFAULT_EVERY_N_ITER = 10


class GcFreezeError(RuntimeError):
    pass


def gc_init_worker(worker_id: int):
    """This function should be called by any forked worker process that uses CUDA.
    It should be called as early as possible in the worker process, ideally in
    the worker_init_fn of the DataLoader.

    By keeping a reference to all CUDA tensors in the worker process, we can
    prevent the forked tensors from being garbage collected."""

    global _frozen_cuda_tensors_initialized, _frozen_cuda_tensors

    num_tensors = 0
    for o in gc.get_objects():
        try:
            if o is not reduce_op:
                if isinstance(o, torch.Tensor):
                    if isinstance(o, ShardedTensorBase) or o.is_cuda:
                        # Calling .is_cuda or any hasattr on ShardedTensor will raise an error
                        # Hence, o.is_cuda is only called if o is not a ShardedTensor (in the if above)

                        _frozen_cuda_tensors.add(o)
                        num_tensors += 1
                elif isinstance(o, torch.utils.data.dataloader._MultiProcessingDataLoaderIter):
                    o._shutdown = True
        except ReferenceError:
            # Can happen if the object is a weakref proxy, don't care
            pass

    _frozen_cuda_tensors_initialized = True


class GcDataset(BaseWrapperDataset[T_sample, T_sample], Generic[T_sample]):
    """Applies a garbage collection step. This is needed, because python garbage collection
    does not work well with very large objects, such as tensors. This case happens, if there are
    a few hundred objects created and released every epoch (some of them being (large) tensors),
    where a lot of them are alive at the same time, but released later. In that case, those objects
    may end up in gc generation 2, where they may live until a lot of objects have been created,
    until automatic garbage collection of gen2 is actually triggered. To avoid this memory leak,
    `gc.collect()` is best to be called regularly. In addition, if `gc.freeze()` is used before the
    loop, it will remove the objects currently alive from garbage collection checks, thus making the
    gc faster.
    """

    every_n_iter: int
    freeze: bool

    def __init__(
        self,
        dataset: SavableDataset[T_sample],
        *,
        worker_config: WorkerConfig,
        every_n_iter: int = GC_DEFAULT_EVERY_N_ITER,
        freeze: bool = True,
    ):
        """Construct a GcDataset, which applies garbage collection every `every_n_iter` iterations.

        Args:
            dataset: The input dataset to wrap
            every_n_iter: How often to perform garbage collection
            freeze: If true, run `gc.freeze()` before the loop, and `gc.unfreeze()` after the loop.
                This will speed up garbage collection, but will keep all initially alive objects
                alive until the end of the loop (i.e. if the dataset state was restored, that state
                will be saved as well).
        """
        super().__init__(dataset, worker_config=worker_config)
        self.every_n_iter = every_n_iter
        self.freeze = freeze

    def reset_state_own(self) -> None:
        return

    def __len__(self):
        return len(self.dataset)

    def __iter__(self) -> Iterator[T_sample]:
        in_worker = torch.utils.data.get_worker_info() is not None
        if in_worker and not _frozen_cuda_tensors_initialized:
            raise GcFreezeError(
                "You are using GcDataset with multiple workers, but forgot to call gc_init_worker() in at least one forked worker process."
            )

        if self.freeze:
            gc.collect()
            gc.freeze()
        try:
            iter = 0
            for sample in self.dataset:
                yield sample
                iter += 1
                if iter >= self.every_n_iter:
                    gc.collect()
                    iter = 0
        finally:
            if self.freeze:
                gc.unfreeze()

    def config(self) -> Dict[str, Any]:
        # This is transparent, no config to be saved (it does not affect the dataset)
        return self.dataset.config()

    def __str__(self):
        return f"GcDataset(every_n_iter={self.every_n_iter}, dataset={self.dataset}, freeze={self.freeze})"



# Copyright (c) 2025, NVIDIA CORPORATION.
# SPDX-License-Identifier: BSD-3-Clause

from typing import (
    Any,
    Callable,
    Dict,
    Generator,
    Generic,
    Iterator,
    Optional,
    Tuple,
    TypeVar,
    Union,
)

from torch.utils.data import IterableDataset

from megatron.energon.errors import SYSTEM_EXCEPTIONS, FatalSampleError
from megatron.energon.flavors.base_dataset import SavableDataset, set_sample_restore_key
from megatron.energon.source_info import SourceInfo
from megatron.energon.worker import WorkerConfig
from megatron.energon.wrappers._log_exception import log_exception
from megatron.energon.wrappers.base import BaseWrapperDataset, SampleIndex, get_sample_restore_key

T_sample = TypeVar("T_sample")
T_sample_out = TypeVar("T_sample_out")


class IterMapDataset(BaseWrapperDataset[T_sample, T_sample_out], Generic[T_sample, T_sample_out]):
    """This dataset wrapper applies a custom function to transform the stream of samples and yield
    a new stream of samples.
    If used in a savable dataset context, it is critical, that `iter_map_fn` is either stateless,
    or that the state of the `iter_map_fn` is saved and restored externally.
    """

    iter_map_fn: Callable[[Iterator[T_sample]], Iterator[T_sample_out]]
    len_map_fn: Callable[[int], int]
    error_handler: Callable[[Exception, Optional[T_sample], list[SourceInfo]], None]
    stateless_iter_fn: bool
    iter_map_fn_config: Optional[Union[Dict[str, Any], Callable[[], Dict[str, Any]]]]
    _sample_index: SampleIndex

    _savable_fields = ("_sample_index",)

    def __init__(
        self,
        dataset: SavableDataset[T_sample],
        iter_map_fn: Callable[[Iterator[T_sample]], Iterator[T_sample_out]],
        *,
        len_map_fn: Callable[[int], int] = lambda x: x,
        error_handler: Callable[
            [Exception, Optional[T_sample], list[SourceInfo]], None
        ] = log_exception,
        stateless_iter_fn: bool = False,
        iter_map_fn_config: Optional[Union[Dict[str, Any], Callable[[], Dict[str, Any]]]] = None,
        worker_config: WorkerConfig,
    ):
        """Construct a IterMapDataset.
        For saving and restoring samples, the iter_map_fn must only yield 0 or 1 sample per
        iterated sample.

        Args:
            dataset: The input dataset to wrap
            iter_map_fn: The function to apply to the stream of samples. Returns a new stream of
                samples. If savability should be preserved, this function should be stateless.
            len_map_fn: The function to apply to the length of the dataset. Returns the new
                (approximate) length of the resulting stream of samples based on the original
                length.
            error_handler: Handler for errors. Defaults to logging and ignoring the exception.
            stateless_iter_fn: If true, assume the iter_map_fn is deterministic and stateless
                (it does not aggregate samples (thus key for random access can propagate to inner
                dataset), yielding zero or multiple samples per fetched sample is fine).
                Defaults to False.
            iter_map_fn_config: Configuration for the iter_map_fn function. If callable, it should return the
                configuration. Defaults to None.
            worker_config: Configuration for the workers.
        """
        super().__init__(dataset, worker_config=worker_config)
        self.iter_map_fn = iter_map_fn
        self.len_map_fn = len_map_fn
        self.error_handler = error_handler
        self.stateless_iter_fn = stateless_iter_fn
        self.iter_map_fn_config = iter_map_fn_config

        self.reset_state_own()

    def reset_state_own(self) -> None:
        self._sample_index = SampleIndex(self.worker_config, src=self)

    def __len__(self):
        return self.len_map_fn(len(self.dataset))

    def __iter__(self) -> Iterator[T_sample_out]:
        last_sample_wrapper = _LastSampleWrapper(self.dataset)
        # The iter_map_fn is stateless. Thus we need to know which inner sample created the
        # outer sample, and the relative outer sample index, so we can restore it.

        # This is the sample index within the currently yielded sample
        iter_idx = 0
        sample_idx = 0
        sample_restore_keys = []

        def reset_idx_iter() -> Generator[T_sample, None, None]:
            # Resets the inner sample index
            nonlocal iter_idx, sample_restore_keys
            for entry in last_sample_wrapper:
                iter_idx = 0
                sample_restore_keys.append(get_sample_restore_key(entry))
                yield entry

        ds_iter = iter(reset_idx_iter())

        # While True will break when the inner dataset is exhausted, but may continue on exception
        while True:
            iter_idx = 0
            try:
                for sample_idx, sample in self._sample_index.iter_ctx(self.iter_map_fn(ds_iter)):
                    yield set_sample_restore_key(
                        sample,
                        sample_idx,
                        iter_idx,
                        *sample_restore_keys,
                        src=self,
                    )
                    sample_restore_keys.clear()
                    iter_idx += 1
            except SYSTEM_EXCEPTIONS:
                raise FatalSampleError.from_sample(last_sample_wrapper.last_sample)
            except Exception as e:
                self.error_handler(e, last_sample_wrapper.last_sample)
            else:
                break

    def can_restore_sample(self) -> bool:
        return super().can_restore_sample() and self.stateless_iter_fn

    def assert_can_restore(self) -> None:
        assert self.stateless_iter_fn, (
            "IterMapDataset can only restore samples if iter_map_fn is stateless."
        )
        super().assert_can_restore()

    def restore_sample(self, restore_key: Tuple[Union[str, int, tuple], ...]) -> T_sample:
        self.assert_can_restore()
        id, sample_idx, iter_idx, *sample_restore_keys = restore_key
        assert id == type(self).__name__
        assert isinstance(iter_idx, int)
        inner_iter = iter(
            self.iter_map_fn(
                (self.dataset.restore_sample(inner_index) for inner_index in sample_restore_keys)
            )
        )
        try:
            # Skip inner yielded samples to get the correct sample
            for skip_idx in range(iter_idx):
                with self._sample_index.ctx(sample_idx - iter_idx + skip_idx):
                    next(inner_iter)
            # This is the sample to restore
            with self._sample_index.ctx(sample_idx):
                sample = next(inner_iter)
            return set_sample_restore_key(
                sample,
                sample_idx,
                iter_idx,
                *sample_restore_keys,
                src=self,
            )
        except StopIteration:
            raise RuntimeError(
                "Generator did not yield enough samples, but is marked stateless/deterministic."
            )
        finally:
            # Properly close if it's a generator
            if hasattr(inner_iter, "close"):
                inner_iter.close()

    def config(self) -> Dict[str, Any]:
        return {
            "type": type(self).__qualname__,
            "dataset": self.dataset.config(),
            "iter_map_fn": self._function_config(self.iter_map_fn),
            **(
                {
                    "iter_map_fn_config": (
                        self.iter_map_fn_config()
                        if callable(self.iter_map_fn_config)
                        else self.iter_map_fn_config
                    )
                }
                if self.iter_map_fn_config
                else {}
            ),
            "len_map_fn": self._function_config(self.len_map_fn),
            "error_handler": self._function_config(self.error_handler),
        }

    def __str__(self):
        return f"IterMapDataset(iter_map_fn={self.iter_map_fn}, dataset={self.dataset})"


class _LastSampleWrapper:
    """
    Wraps the inner dataset and stores the last iterated sample.
    """

    last_sample: Optional[T_sample] = None
    dataset: IterableDataset[T_sample]

    def __init__(self, dataset: IterableDataset[T_sample]):
        self.dataset = dataset

    def __iter__(self) -> Iterator[T_sample]:
        for sample in self.dataset:
            self.last_sample = sample
            yield sample



# Copyright (c) 2025, NVIDIA CORPORATION.
# SPDX-License-Identifier: BSD-3-Clause

import hashlib
import random
from typing import Any, List, Mapping, Optional, Sequence, TypeVar

import numpy
import torch
import torch.distributed
import torch.utils.data

from megatron.energon.edataclass import edataclass
from megatron.energon.savable import FlexState, Savable
from megatron.energon.worker import WorkerConfig

T = TypeVar("T")


class WorkerRng(Savable):
    """Helper class for getting a worker random generator, which is still in itself deterministic.
    If not in a worker, uses the global random generator's seed to initialize a new rng."""

    worker_config: WorkerConfig

    _rng: Optional[torch.Generator] = None

    _restore_state: Optional[bytes] = None

    def __init__(self, worker_config: WorkerConfig):
        self.worker_config = worker_config

    @property
    def rng(self) -> torch.Generator:
        if self._rng is None or self._restore_state is not None:
            self.worker_config.assert_worker()
            self._rng = torch.Generator()
            if self._restore_state is not None:
                self._rng.set_state(
                    torch.frombuffer(
                        bytearray(self._restore_state),
                        dtype=torch.uint8,
                    ).clone()
                )
            else:
                # Restore to initial state (either due to zero sized states, or just initial state)
                self._rng.manual_seed(self.worker_config.worker_seed())
            self._restore_state = None
        return self._rng

    def randbelow(self, n: int) -> int:
        return torch.randint(0, n, (), generator=self.rng).item()

    def choice_idx(self, probs: torch.Tensor) -> int:
        if len(probs) == 1:
            return 0
        else:
            # Custom implementation of multinomial to ensure consistency
            # Torch changed their implementation of torch.multinomial in 2.7.0 and to be
            # consistent with any torch version, we use a custom implementation here instead.
            # This is anyways just a very simple case of multinomial, thus this should be fine.
            # Actually, benchmarks show that this is faster than torch.multinomial by a factor of
            # 10 even on CPU.
            cdf = torch.cumsum(probs, dim=0)
            val = torch.rand(1, generator=self.rng) * cdf[-1]
            return torch.searchsorted(cdf, val).item()

    def choice(self, l: List[T], probs: Optional[torch.Tensor] = None) -> T:
        if probs is None:
            return l[self.randbelow(len(l))]
        assert len(l) == len(probs)
        return l[self.choice_idx(probs)]

    def shuffle(self, l: List[T]) -> List[T]:
        """Returns a new list with shuffled entries"""
        p = torch.randperm(len(l), generator=self.rng)
        return [l[p[i]] for i in range(len(l))]

    def rand_pop(self, l: List[T]) -> T:
        return l.pop(self.randbelow(len(l)))

    def save_state(self) -> FlexState:
        return FlexState(rng=None if self.rng is None else bytes(self.rng.get_state().tolist()))

    def restore_state(self, state: FlexState):
        if state["rng"] is None:
            self._restore_state = None
        else:
            self._restore_state = state["rng"]


@edataclass
class SystemRngState:
    """The state of the global random generators.

    Note that the data types of the internal RNG states are implementation details of the
    respective libraries and may change in the future.

    Python does not even specify the type in their docs. Hence we will allow arbitrary types,
    because all that matters is that we can save and restore them. We will not use the data
    anywhere else.
    """

    torch: Any  # Currently `torch.Tensor`
    numpy: Any  # Currently `dict[str, Any] | tuple[str, NDArray[uint32], int, int, float]`
    random: Any  # Currently a nested tuple

    def _hashable_value(self, value: Any) -> Any:
        if isinstance(value, (int, float, bool, str)) or value is None:
            return value
        elif isinstance(value, torch.Tensor):
            return self._hashable_value(value.tolist())
        elif isinstance(value, numpy.ndarray):
            return self._hashable_value(value.tolist())
        elif isinstance(value, Mapping):
            return tuple(
                (self._hashable_value(k), self._hashable_value(v)) for k, v in value.items()
            )
        elif isinstance(value, Sequence):
            return tuple(self._hashable_value(v) for v in value)
        else:
            raise ValueError(f"Cannot hash value of type {type(value)}: {value!r}")

    def __repr__(self):
        # If the hash is the same, the state is the same. Should suffice to identify the state.
        return f"SystemRngState(hash={hash(self._hashable_value((self.torch, self.numpy, self.random)))})"


class SystemRng:
    """A class to seed, save or restore the global random generators.
    This affects torch, numpy and the standard library random module."""

    @staticmethod
    def seed(seed: int) -> None:
        """Seeds the global random generators."""
        torch.manual_seed(seed)
        numpy.random.seed(seed)
        random.seed(seed)

    @staticmethod
    def save_state() -> SystemRngState:
        """Saves the global rng state for torch, numpy and random."""
        return SystemRngState(
            torch=torch.get_rng_state(),
            numpy=numpy.random.get_state(),
            random=random.getstate(),
        )

    @staticmethod
    def restore_state(state: SystemRngState) -> None:
        """Restores the global rng state for torch, numpy and random."""
        torch.set_rng_state(state.torch)
        numpy.random.set_state(state.numpy)
        random.setstate(state.random)

    @staticmethod
    def get_seed_from_args(*args: Any) -> int:
        """Deterministically generates a seed from the given arguments.
        The str() representation of each arg is used."""

        # Use a deterministic hash function to compute the seed
        hash_digest = hashlib.sha1("|".join([str(obj) for obj in args]).encode("utf-8")).digest()

        # We use the first 4 bytes of the hash as the seed and fix the endianness
        seed_value = int.from_bytes(hash_digest[:4], byteorder="big")

        return seed_value

    @staticmethod
    def seed_args(*args: Any) -> None:
        """Seeds the global random generators deterministically from the given arguments."""
        SystemRng.seed(SystemRng.get_seed_from_args(*args))



# Copyright (c) 2025, NVIDIA CORPORATION.
# SPDX-License-Identifier: BSD-3-Clause

from megatron.energon.task_encoder.base import (
    AugmentTaskEncoder,
    Batch,
    DefaultTaskEncoder,
    TaskEncoder,
    batch_list,
    batch_pad_stack,
    batch_stack,
    generic_batch,
    stateless,
)
from megatron.energon.task_encoder.cooking import Cooker, basic_sample_keys, cooker
from megatron.energon.task_encoder.loader import (
    get_train_dataset,
    get_val_dataset,
    get_val_datasets,
)

__all__ = [
    "AugmentTaskEncoder",
    "basic_sample_keys",
    "batch_list",
    "batch_pad_stack",
    "batch_stack",
    "Batch",
    "cooker",
    "Cooker",
    "DefaultTaskEncoder",
    "generic_batch",
    "get_train_dataset",
    "get_val_dataset",
    "get_val_datasets",
    "stateless",
    "TaskEncoder",
]



# Copyright (c) 2025, NVIDIA CORPORATION.
# SPDX-License-Identifier: BSD-3-Clause

import gc
import multiprocessing
import queue
import threading
import time
import traceback
import weakref
from functools import partial
from typing import (
    Any,
    ClassVar,
    Dict,
    Generic,
    Iterator,
    List,
    Optional,
    Sequence,
    Tuple,
    TypeVar,
    Union,
)

import numpy as np
import torch
import torch.distributed
import torch.multiprocessing
from torch.utils.data import DataLoader, IterableDataset

from megatron.energon.cache import CachePool
from megatron.energon.edataclass import edataclass
from megatron.energon.errors import deprecated
from megatron.energon.flavors.base_dataset import (
    FlexState,
    SavableDataset,
    State,
    add_sample_restore_key,
)
from megatron.energon.rng import SystemRng, SystemRngState
from megatron.energon.worker import WorkerConfig
from megatron.energon.wrappers.base import BaseWrapperDataset
from megatron.energon.wrappers.batch_dataset import BatchDataset
from megatron.energon.wrappers.gc_dataset import GC_DEFAULT_EVERY_N_ITER, GcDataset, gc_init_worker
from megatron.energon.wrappers.log_sample_dataset import default_get_keys
from megatron.energon.wrappers.watchdog_dataset import WatchdogDataset

T = TypeVar("T")


def _init_worker(seed_per_worker: List[int], worker_id: int):
    """Initializes the the worker process.

    Sets the random seeds and prepare EPath for the forked worker process.
    """
    gc_init_worker(worker_id)

    worker_seed = seed_per_worker[worker_id]

    SystemRng.seed(worker_seed)


class SimpleSavableDatasetWrapper(BaseWrapperDataset[T, Tuple[int, int, T]], Generic[T]):
    """Wrapper for non-multiprocessing savable datasets. Restarts the inner dataset. This class is
    not intended to be used directly."""

    #: The cache pool to use for the dataset.
    cache_pool: CachePool

    _state_restored: bool
    _sample_index: int

    _savable_fields = ("_sample_index",)

    def __init__(
        self, dataset: SavableDataset[T], worker_config: WorkerConfig, cache_pool: CachePool
    ):
        """
        Args:
            dataset: The dataset to wrap.
            worker_config: The worker config to use for the dataset.
            cache_pool: The cache pool to use for the dataset.
        """
        super().__init__(dataset, worker_config=worker_config)
        self.cache_pool = cache_pool

        self.reset_state_own()

    def reset_state_own(self) -> None:
        self._sample_index = 0
        self._state_restored = False

    def __len__(self):
        return len(self.dataset)

    def __iter__(self):
        self._state_restored = True
        worker_id = self.worker_config.rank_worker_id()
        global_worker_id = self.worker_config.global_worker_id()
        while self._state_restored:
            self._state_restored = False
            self.worker_config.worker_activate(self._sample_index, cache_pool=self.cache_pool)
            worker_active = True
            try:
                for src_data in self.dataset:
                    self.worker_config.worker_deactivate()
                    worker_active = False
                    sample_index = self._sample_index
                    src_data = add_sample_restore_key(
                        src_data, global_worker_id, sample_index, src=self
                    )
                    self._sample_index += 1
                    yield worker_id, sample_index, src_data
                    if self._state_restored:
                        # Restart iterator after restore
                        break
                    self.worker_config.worker_activate(
                        self._sample_index, cache_pool=self.cache_pool
                    )
                    worker_active = True
            finally:
                if worker_active:
                    self.worker_config.worker_deactivate()

    def restore_sample(self, restore_key: Tuple[Union[str, int, tuple], ...]) -> T:
        id, global_worker_id, sample_idx = restore_key[:3]
        assert id == type(self).__name__
        restore_key = restore_key[3:]
        self.worker_config.worker_activate(
            sample_idx, override_global_rank=global_worker_id, cache_pool=self.cache_pool
        )
        try:
            return add_sample_restore_key(
                self.dataset.restore_sample(restore_key),
                global_worker_id,
                sample_idx,
                src=self,
            )
        finally:
            self.worker_config.worker_deactivate()

    def config(self) -> Dict[str, Any]:
        return self.dataset.config()

    def __str__(self):
        return f"SimpleSavableDatasetWrapper(dataset={self.dataset})"


@edataclass
class SavableDatasetState(State):
    """State of the dataset wrapper. It stores the global random states and the index of the next
    sample to be returned from the dataset. This class is not intended to be used directly, but by
    :class:`megatron.energon.SavableDatasetWrapper`."""

    #: The state of all the system random number generators
    rng: SystemRngState
    #: The state of the savable dataset
    dataset_state: FlexState
    #: Index of the next sample to be returned from the dataset
    sample_index: int

    def __repr__(self):
        return f"SavableDatasetState(rng={self.rng!r}, sample_index={self.sample_index})"


@edataclass
class SavableCheckpoint:
    """Checkpoint data for :class:`megatron.energon.SavableDatasetWrapper`. An instance is created
    regularly to be able to save the state of the dataset wrapper before the currently emitted
    sample.
    """

    #: The state of the wrapper
    state: Optional[SavableDatasetState]
    #: The time at which the checkpoint was created
    checkpoint_time: float
    #: Index of the next sample to be returned from the dataset after restoring the checkpoint
    sample_index: int


@edataclass
class SavableDatasetCheckpoint(State):
    """Checkpoint data for :class:`megatron.energon.SavableDatasetWrapper`. The checkpoint state
    represents a state before that checkpoint, with an offset (i.e. samples to be skipped)."""

    #: The state of the wrapper at the sample index when the checkpoint was created.
    state: Optional[SavableDatasetState]
    #: Offset of the checkpoint to the actual sample index to be restored.
    offset: int


class SavableDatasetWrapper(IterableDataset[Tuple[int, int, T]], Generic[T]):
    """Internal class for wrapping a savable dataset for a worker process. Provides communication
    with the :class:`megatron.energon.SavableDataLoader`. This class is not intended to be used directly.
    See :class:`megatron.energon.SavableDataLoader` for more information."""

    #: The wrapped dataset
    dataset: SavableDataset[T]
    #: The configuration of the worker process
    worker_config: WorkerConfig
    #: The time interval in seconds to wait at minimum between two checkpoints
    checkpoint_every_sec: float
    #: The minimum number of samples to be emitted between two checkpoints. Should be `number of
    # workers * 2`.
    checkpoint_every_min_n_samples: int
    #: The number of checkpoints to keep in memory, before discarding. Should be 2.
    n_checkpoints: int
    #: The cache pool to use for the dataset.
    cache_pool: CachePool
    #: The queue of the worker process to receive commands from the `SavableDataLoader`.
    _cmd_queues: List[torch.multiprocessing.Queue]
    #: The queue of the worker process to send results to the `SavableDataLoader`.
    _result_queues: List[torch.multiprocessing.Queue]

    _sample_index: int = 0
    _worker_offset: int = 0
    _last_checkpoints: List[SavableCheckpoint]

    _workers_restore_from: List[Optional[SavableDatasetState]] = list()
    _workers_skip_samples: List[int]

    _running: bool = False
    _command_lock: Optional[threading.RLock] = None
    _cmd_thread: Optional[threading.Thread] = None

    def __init__(
        self,
        dataset: SavableDataset[T],
        worker_config: WorkerConfig,
        checkpoint_every_sec: float,
        checkpoint_every_min_n_samples: int,
        n_checkpoints: int = 2,
        *,
        cmd_queues: List[torch.multiprocessing.Queue],
        result_queues: List[torch.multiprocessing.Queue],
        cache_pool: CachePool,
    ):
        """
        Create the savable dataset wrapper for multiprocessing data loading.

        Args:
            dataset: The dataset to wrap
            worker_config: The worker config as used by all datasets
            checkpoint_every_sec: The time interval in seconds to wait at minimum between two
                checkpoints.
            checkpoint_every_min_n_samples: The minimum number of samples to be emitted between
                two checkpoints. Should be `number of workers * 2`.
            n_checkpoints: Number of checkpoints to keep.
            cmd_queues: The command queues for communicating with the worker processes.
            result_queues: The result queues for communicating with the worker processes.
            cache_pool: The cache pool to use for the dataset.
        """
        num_workers = max(worker_config.num_workers, 1)

        self.dataset = dataset
        self.worker_config = worker_config
        self.checkpoint_every_sec = checkpoint_every_sec
        self.checkpoint_every_min_n_samples = checkpoint_every_min_n_samples
        self.n_checkpoints = n_checkpoints
        self._last_checkpoints = [
            SavableCheckpoint(state=None, checkpoint_time=time.perf_counter(), sample_index=-1)
        ]
        self._workers_restore_from = [None] * num_workers
        self._workers_skip_samples = [0] * num_workers
        self._cmd_queues = cmd_queues
        self._result_queues = result_queues
        self.cache_pool = cache_pool

    @staticmethod
    def _command_thread(self: "SavableDatasetWrapper"):
        """The internal thread, which processes the command and result queues. This thread is
        static, because `self` is actually passed as weakref proxy, to avoid keeping the dataset
        alive via the thread.
        """
        # print(f"{id(self)}:{multiprocessing.current_process().ident} Worker command thread starting")
        assert self._command_lock is not None

        try:
            while self._running:
                try:
                    cmd_args = self._cmd_queues[self._worker_id].get(timeout=1)
                except queue.Empty:
                    continue
                # print(f"recv cmd {cmd_args}")
                with self._command_lock:
                    cmd = cmd_args[0]
                    if cmd is None:
                        break
                    try:
                        fn = getattr(self, cmd)
                        self._result_queues[self._worker_id].put(
                            {self._worker_id: fn(*cmd_args[1:])}
                        )
                        # print(f"result sent")
                    except Exception as e:
                        traceback.print_exc()
                        self._result_queues[self._worker_id].put({self._worker_id: e})
                        # print(f"exc sent")
        except BaseException:
            traceback.print_exc()
            raise
        finally:
            pass
            # print(f"{id(self)}:{multiprocessing.current_process().ident} Worker command thread closing")

    def __len__(self):
        return len(self.dataset)

    def __del__(self):
        if self._cmd_thread is not None:
            # print(f"{id(self)}:{multiprocessing.current_process().ident} Closing cmd thread")
            self._running = False
            self._cmd_thread.join()
            self._command_lock = None
            self._cmd_thread = None
            # print(f"{id(self)}:{multiprocessing.current_process().ident} Cmd thread closed")

    def __iter__(self):
        # First: Set the worker offset globally for the current worker
        WorkerConfig.worker_id_offset = self._worker_offset
        self._worker_id = self.worker_config.rank_worker_id()
        global_worker_id = self.worker_config.global_worker_id()
        if self._cmd_thread is None:
            self._running = True
            self._command_lock = threading.RLock()
            weakref_self = weakref.proxy(self)
            self._cmd_thread = threading.Thread(
                target=SavableDatasetWrapper._command_thread,
                name="command_thread",
                args=(weakref_self,),
                daemon=True,
            )
            self._cmd_thread.start()
            # atexit.register(lambda: weakref_self.__del__())
        try:
            assert self._command_lock is not None
            with self._command_lock:
                if self._workers_restore_from[self._worker_id] is not None:
                    my_state = self._workers_restore_from[self._worker_id]
                    my_ds_state = my_state.dataset_state
                    assert my_state is not None
                    if my_ds_state is None:
                        self.dataset.reset_state_deep()
                    else:
                        self.dataset.restore_state(my_ds_state)
                    self._restore_state(my_state)
                    self._workers_restore_from[self._worker_id] = None
                else:
                    # Store the initial state of the worker if we stop before the first sample
                    self._store_checkpoint()
                # If skipping, also restart the iterator to reach the start of the restored
                # checkpoint
                last_was_skip = True
                while last_was_skip:
                    dataset_has_samples = False
                    self.worker_config.worker_activate(
                        self._sample_index, cache_pool=self.cache_pool
                    )
                    worker_active = True
                    try:
                        for src_data in self.dataset:
                            self.worker_config.worker_deactivate()
                            worker_active = False
                            dataset_has_samples = True
                            if self._workers_skip_samples[self._worker_id] > 0:
                                # Skip ahead to reach the start of the restored checkpoint
                                # print(f"Skip [{self._sample_index}:{self._worker_id}] {src_data}")
                                self._workers_skip_samples[self._worker_id] -= 1
                                self._sample_index += 1
                                last_was_skip = True
                                self.worker_config.worker_activate(
                                    self._sample_index, cache_pool=self.cache_pool
                                )
                                worker_active = True
                                continue
                            last_was_skip = False
                            sample_index = self._sample_index
                            add_sample_restore_key(
                                src_data, global_worker_id, sample_index, src=self
                            )
                            self._sample_index += 1
                            self._store_checkpoint()
                            try:
                                self._command_lock.release()
                                # print(f"{id(self)}:{multiprocessing.current_process().ident} Lock released")
                                # Commands may be executed only when data was yielded, not during
                                # iteration fetching.
                                # print(f"Yield next data [{sample_index}:{self._worker_id}] {src_data}")
                                yield self._worker_id, sample_index, src_data
                            finally:
                                # print(f"{id(self)}:{multiprocessing.current_process().ident} Lock acquiring")
                                self._command_lock.acquire()
                                # print(f"{id(self)}:{multiprocessing.current_process().ident} Lock acquired")
                            self.worker_config.worker_activate(
                                self._sample_index, cache_pool=self.cache_pool
                            )
                            worker_active = True
                    finally:
                        if worker_active:
                            self.worker_config.worker_deactivate()

                    # If the dataset is empty, don't try again and again
                    if not dataset_has_samples:
                        break
        finally:
            # print(f"{id(self)}:{multiprocessing.current_process().ident} Worker iter closing")
            # Always store a final checkpoint (it's likely to be saved)
            self._store_checkpoint(force=True)

    def _store_checkpoint(self, force: bool = False) -> None:
        """
        Internally create a checkpoint for the current state. This is required to store states
        from the past, which have already been yielded here, but not yet been retrieved from the
        intermediate queues.

        Args:
            force: If true, ignore time or frequency condition.
        """
        if (
            force
            or (
                self._last_checkpoints[-1].checkpoint_time + self.checkpoint_every_sec
                < time.perf_counter()
                and self._last_checkpoints[-1].sample_index + self.checkpoint_every_min_n_samples
                <= self._sample_index
            )
            or self._sample_index <= 1
        ):
            # print(f"Storing checkpoint at {self._worker_id}:{self._sample_index}")
            self._last_checkpoints.append(
                SavableCheckpoint(
                    state=self._save_state(),
                    checkpoint_time=time.perf_counter(),
                    sample_index=self._sample_index,
                )
            )
            if len(self._last_checkpoints) > self.n_checkpoints:
                self._last_checkpoints.pop(0)

    def _save_state(self) -> SavableDatasetState:
        """Saves the internal state"""
        (
            np_tp,
            np_state,
            pos,
            has_gauss,
            cached_gaussian,
        ) = np.random.get_state()
        return SavableDatasetState(
            rng=SystemRng.save_state(),
            dataset_state=self.dataset.save_state(),
            sample_index=self._sample_index,
        )

    def _restore_state(self, state: SavableDatasetState) -> None:
        """Restores the internal worker state"""
        assert torch.utils.data.get_worker_info() is not None, "Can only restore in worker process"
        if state.rng is None:
            SystemRng.seed(torch.initial_seed() & 0xFFFFFFFF)
        else:
            SystemRng.restore_state(state.rng)

        self._sample_index = state.sample_index
        self._last_checkpoints = [
            SavableCheckpoint(
                state=self._save_state(),
                checkpoint_time=time.perf_counter(),
                sample_index=self._sample_index,
            )
        ]

    def get_checkpoint(self, last_sample_indexes: List[int]) -> SavableDatasetCheckpoint:
        """
        Get a checkpoint given the last emitted sample indexes for all workers.

        Args:
            last_sample_indexes: The last emitted sample indexes for all workers.

        Returns:
            The found checkpoint including the offset to the next sample index
        """
        sample_index = last_sample_indexes[self._worker_id] + 1
        for checkpoint in reversed(self._last_checkpoints):
            if checkpoint.sample_index <= sample_index:
                # print(f"Found cp for {sample_index} at {checkpoint.sample_index}")
                return SavableDatasetCheckpoint(
                    state=checkpoint.state,
                    offset=sample_index - checkpoint.sample_index,
                )

        # Immediate save after restore
        if len(self._last_checkpoints) == 0:
            return SavableDatasetCheckpoint(
                state=self._workers_restore_from[self._worker_id],
                offset=self._workers_skip_samples[self._worker_id],
            )
        raise ValueError("No checkpoint found")

    def restore_checkpoint(
        self,
        worker_states: Optional[List[SavableDatasetCheckpoint]],
        worker_offset: int,
    ) -> None:
        """
        Restores the merged checkpoint from all worker processes.

        Args:
            worker_states: The state to restore for each worker
            worker_offset: The offset of the last worker which has emitted a sample. This will be
                set in all worker processes to ensure the right worker starts as first.
        """
        assert torch.utils.data.get_worker_info() is None, "Cannot restore in worker process"
        num_workers = max(self.worker_config.num_workers, 1)

        if worker_states is None:
            self._workers_restore_from = [None] * num_workers
            assert worker_offset == 0
            self._worker_offset = 0
            self._workers_skip_samples = [0] * num_workers
        else:
            assert isinstance(worker_states, list)
            assert len(worker_states) == num_workers
            assert isinstance(worker_states[0], SavableDatasetCheckpoint)

            self._worker_offset = worker_offset

            # Tear the state_list apart (which has len=num_workers)
            # and store the states in the internal arrays
            self._workers_restore_from = [state.state for state in worker_states]
            self._workers_skip_samples = [state.offset for state in worker_states]

    def get_initial_checkpoint(self) -> Optional[List[SavableDatasetCheckpoint]]:
        """
        Get the initial checkpoint for all worker processes if they have not started yet.

        Returns:
            The initial checkpoint for all worker processes and the worker offset.
        """
        assert torch.utils.data.get_worker_info() is None, (
            "Cannot get initial checkpoint in worker process"
        )
        if all(s is None for s in self._workers_restore_from):
            assert all(s == 0 for s in self._workers_skip_samples)
            # Initial state, no checkpoint
            return None

        return [
            SavableDatasetCheckpoint(
                state=state,
                offset=offset,
            )
            for state, offset in zip(self._workers_restore_from, self._workers_skip_samples)
        ]

    def can_restore_sample(self) -> bool:
        return self.dataset.can_restore_sample()

    def restore_sample(self, restore_key: Tuple[Union[str, int, tuple], ...]) -> T:
        id, global_worker_id, sample_idx = restore_key[:3]
        assert id == type(self).__name__
        restore_key = restore_key[3:]
        self.worker_config.worker_activate(sample_idx, override_global_rank=global_worker_id)
        try:
            return add_sample_restore_key(
                self.dataset.restore_sample(restore_key),
                global_worker_id,
                sample_idx,
                src=self,
            )
        finally:
            self.worker_config.worker_deactivate()

    def config(self) -> Dict[str, Any]:
        return self.dataset.config()

    def __str__(self):
        return f"SavableDatasetWrapper(dataset={self.dataset})"


@edataclass
class SavableDataLoaderState(State):
    """Saved state of the :class:`megatron.energon.SavableDataLoader`. Contains the state for all worker
    processed of a single rank."""

    #: The internal state of the dataset (for each worker process)
    worker_states: List[Union[SavableDatasetCheckpoint, FlexState]]
    #: Which worker will be the next to emit a sample. Used to restore the proper order
    next_worker_id: int

    #: The micro batch size that was used, if available.
    #: On restore, this is used to compare the new and old micro batch size.
    micro_batch_size: Optional[int]


class SavableDataLoader(DataLoader[T], Generic[T]):
    """DataLoader that supports saving and restoring the state of the dataset.
    When restoring, the dataloader and dataset must be instantiated with the exactly same
    parameters.

    How this works (for no worker processes)
    ----------------------------------------

    1. The state of the dataset is saved using :meth:`megatron.energon.SavableDataset.save_state`
    2. (for compatibility) The state of the dataset is converted to using inner arrays using
       :meth:`megatron.energon.SavableDataset.merge_states`.
    3. The state can be restored using :meth:`megatron.energon.SavableDataset.restore_state` given the
       previously saved (and merged) state.

    How this works (for worker processes)
    -------------------------------------

    - First issue is, that worker processes work with internal queues between processes to pass
      loaded samples to the main process (also to perform collating). This means that the whole
      state of the dataset is not directly accessible from the main process.
    - To solve this issue, the dataset regularly saves a checkpoint of its state to be able to
      resume from that state (and skip the samples that have already been yielded).
    - To have a consistent state, the sample index from the latest yielded samples is saved for all
      worker instances. Thus, the main process knows exactly which sample indexes should come next
      from which worker.
    - Internally, pytorch iterates through the workers in order to retrieve the next worker's
      samples. Unfortunately, that next worker index cannot be restored in pytorch's dataloader,
      thus the workers are shifted internally by that offset
      (see :attr:`megatron.energon.WorkerConfig.worker_id_offset`).

    1. The dataset is wrapped in a :class:`megatron.energon.SavableDatasetWrapper`. This allows the main
       process to communicate with the worker and send commands to the workers and retrieve the
       results.
    2. The state of the dataset is saved using
       :meth:`megatron.energon.SavableDatasetWrapper.get_checkpoint`. This gives the last checkpoint
       from the requested sample index and stores the offset (i.e. number of samples to skip) from
       that checkpoint.
    3. The state is merged using :meth:`megatron.energon.SavableDatasetWrapper.merge_checkpoints`. This
       merges the states of all workers and returns a single state that can be used to restore the
       state of the dataset.
    4. The state can be restored using :meth:`megatron.energon.SavableDatasetWrapper.restore_state`
       before a worker is started, such that all workers initially receive the same state array.
       The worker firstly sets the worker index offset, then uses its (shifted) own index to get its
       required state from the merged state array.

    """

    #: The worker config
    worker_config: WorkerConfig
    #: The wrapped dataset. For multiprocessing, this is a :class:`megatron.energon.SavableDatasetWrapper`
    dataset: Union[SavableDatasetWrapper[T], SimpleSavableDatasetWrapper[T]]

    #: The global ID counter
    _next_id: ClassVar[int] = 0
    #: Class instance id
    id: int = 0

    #: The queues used to send commands to the workers
    cmd_queues: List[torch.multiprocessing.Queue]
    #: The queues used to receive results from the workers
    result_queues: List[torch.multiprocessing.Queue]

    #: Instance of the current data iterator. There shall be only one active iterator, such that the
    # dataset is not iterated multiple times in parallel. The state will proceed.
    _persistent_iterator: Optional[Iterator[T]] = None
    #: The index of the current worker. -1 if not started yet.
    _worker_sample_counters: List[int]
    #: Id of the next worker to retrieve data from
    _next_worker_id: int = 0
    #: Global index of the last yielded sample
    _global_sample_idx: int = 0
    #: Current iterator index of the last yielded sample
    _sample_idx: int = 0

    def __init__(
        self,
        dataset: SavableDataset[T],
        *,
        checkpoint_every_sec: float = 60,
        checkpoint_every_min_n_samples: Optional[int] = None,
        n_checkpoints: Optional[int] = None,
        gc_collect_every_n_steps: int = GC_DEFAULT_EVERY_N_ITER,
        gc_freeze_at_start: bool = True,
        prefetch_factor: int = 2,
        cache_pool: Optional[CachePool] = None,
        watchdog_timeout_seconds: Optional[float] = 60,
        watchdog_initial_timeout_seconds: Optional[float] = None,
        fail_on_timeout: bool = False,
    ):
        """
        Create the dataloader supporting saving and restoring the state.

        Args:
            dataset: The dataset to load.
            worker_config: The worker config to use
            checkpoint_every_sec: This is the time in seconds after which a checkpoint is saved.
                It may take the same duration to restore a checkpoint, but introduces additional
                overhead during reading data from the dataset, so this should be chosen accordingly.
                Only applies if using workers.
            checkpoint_every_min_n_samples: Overwrites the minimum number of samples between
                checkpoints. Defaults to `number of workers * 2`. Only applies if using workers.
            n_checkpoints: The number of checkpoints to keep in memory. Only applies if using
                workers. If None, computes a suitable value.
            gc_collect_every_n_steps: The number of steps after which the garbage collector is
                called. As we're usually handling large (but few) tensors here, and the python
                garbage collection is already full of objects just by importing, this can improve
                the memory footprint quite a lot, and may even be necessary to avoid memory
                overflow.
            gc_freeze_at_start: If true, the garbage collector is frozen at the start of the worker
                processes. This improves the garbage collection performance by a lot.
                In rare cases, this may cause issues and can be disabled. Keep enabled if you
                experience no issues.
            cache_pool: If set, the cache pool to use for the dataset.
            watchdog_timeout_seconds: The timeout in seconds. If None, the watchdog is disabled.
            watchdog_initial_timeout_seconds: The initial timeout in seconds. If None, the timeout is the same as watchdog_timeout_seconds.
            fail_on_timeout: If True, stops the whole process upon timeout, after printing a stack trace.
        """
        self.worker_config = dataset.worker_config
        self.id = self.next_id()

        dataset = WatchdogDataset(
            dataset,
            worker_config=self.worker_config,
            timeout_seconds=watchdog_timeout_seconds,
            initial_timeout_seconds=watchdog_initial_timeout_seconds,
            fail_on_timeout=fail_on_timeout,
        )

        if gc_collect_every_n_steps > 0:
            dataset = GcDataset(
                dataset,
                worker_config=self.worker_config,
                every_n_iter=gc_collect_every_n_steps,
                freeze=gc_freeze_at_start,
            )

        self.cmd_queues = [multiprocessing.Queue() for _ in range(self.worker_config.num_workers)]
        self.result_queues = [
            multiprocessing.Queue() for _ in range(self.worker_config.num_workers)
        ]

        num_procs = max(self.worker_config.num_workers, 1)

        if n_checkpoints is None:
            n_checkpoints = prefetch_factor * num_procs + 1

        if self.worker_config.num_workers > 0:
            if checkpoint_every_min_n_samples is None:
                checkpoint_every_min_n_samples = self.worker_config.num_workers * 2

            dataset = SavableDatasetWrapper(
                dataset,
                self.worker_config,
                checkpoint_every_sec=checkpoint_every_sec,
                checkpoint_every_min_n_samples=checkpoint_every_min_n_samples,
                n_checkpoints=n_checkpoints,
                cmd_queues=self.cmd_queues,
                result_queues=self.result_queues,
                cache_pool=cache_pool,
            )
        else:
            dataset = SimpleSavableDatasetWrapper(
                dataset, self.worker_config, cache_pool=cache_pool
            )

        self._worker_sample_counters = [-1] * num_procs

        kwargs = {}
        if self.worker_config.num_workers > 0:
            kwargs["persistent_workers"] = True
            kwargs["prefetch_factor"] = prefetch_factor

        # Assert that prefetch_factor works well with num_checkpoints.
        # This ensures that the oldest checkpoint is old enough to cover
        # all the buffered samples in the torch dataloader.
        assert prefetch_factor * num_procs + 1 <= n_checkpoints, (
            "When increasing prefetch_factor, also increase n_checkpoints, so that "
            "the number of checkpoints is at least as large as num_workers * prefetch_factor + 1"
        )

        # Compute seeds for each worker, based on current rank
        seed_per_worker = [
            self.worker_config.worker_seed(i) for i in range(self.worker_config.num_workers)
        ]

        super().__init__(
            dataset,
            batch_size=None,
            shuffle=False,
            num_workers=self.worker_config.num_workers,
            pin_memory=True,
            worker_init_fn=partial(_init_worker, seed_per_worker),
            **kwargs,
        )

        if self.worker_config.should_log(level=1):
            self.worker_config.worker_log(
                {
                    "t": "SavableLoader.__init__",
                    "r": self.worker_config.rank,
                    "w": None,
                    "id": self.id,
                    "config": dataset.config(),
                }
            )

    @staticmethod
    def next_id() -> int:
        next_id = SavableDataLoader._next_id
        SavableDataLoader._next_id += 1
        return next_id

    def __iter__(self):
        outerself = self

        class InnerIterator:
            """Internal class which keeps the iterator alive across multiple `iter()` calls.
            If the inner iterator is exhausted, will also exhaust and a new instance is needed.
            Also saves the last sample index and the next worker id.
            """

            finished: bool = False
            iter_idx: int = 0
            id: int

            def __init__(self, iterator):
                self._iterator = iterator
                self.id = outerself.next_id()
                if outerself.worker_config.should_log(level=1):
                    outerself.worker_config.worker_log(
                        {
                            "t": "SavableDataLoader.iter",
                            "r": outerself.worker_config.rank,
                            "w": None,
                            "id": outerself.id,
                            "iter_id": self.id,
                        }
                    )

                # self._debugf = open(
                #     f"worker_samples_rank{outerself.worker_config.rank:02}_t{int(time.time())}.log", "w"
                # )

            def __iter__(self):
                return self

            def __next__(self):
                try:
                    worker_id, sample_idx, sample = next(self._iterator)
                    outerself._worker_sample_counters[worker_id] = sample_idx
                    # If the next sample will be from the first worker, we can safely resume
                    outerself._next_worker_id = (worker_id + 1) % max(outerself.num_workers, 1)
                    # self._debugf.write(
                    #     f"[w={worker_id}, s={sample_idx}] {self._sample_str(sample)}\n"
                    # )
                    # self._debugf.flush()
                    if outerself.worker_config.should_log(level=1):
                        keys = default_get_keys(sample)
                        outerself.worker_config.worker_log(
                            {
                                **{
                                    "t": "SavableDataLoader.yield",
                                    "r": outerself.worker_config.rank,
                                    "w": None,
                                    "id": outerself.id,
                                    "iter_id": self.id,
                                    "worker_id": worker_id,
                                    "worker_idx": sample_idx,
                                    "idx": outerself._sample_idx,
                                    "iter_idx": self.iter_idx,
                                    "global_idx": outerself._global_sample_idx,
                                },
                                **({} if keys is None else {"keys": keys}),
                            }
                        )
                    outerself._sample_idx += 1
                    outerself._global_sample_idx += 1
                    self.iter_idx += 1
                    return sample
                except StopIteration:
                    self.finished = True
                    outerself._next_worker_id = 0
                    if outerself.worker_config.should_log(level=1):
                        outerself.worker_config.worker_log(
                            {
                                "t": "SavableDataLoader.StopIteration",
                                "r": outerself.worker_config.rank,
                                "w": None,
                                "id": outerself.id,
                                "iter_id": self.id,
                            }
                        )
                    raise

        if self.num_workers > 0:
            # Always keep same iterator alive, as long as it yields data
            if self._persistent_iterator is None or self._persistent_iterator.finished:
                self._persistent_iterator = InnerIterator(super().__iter__())
                self._sample_idx = 0
                # print("New Iterator", self._persistent_iterator)
            return self._persistent_iterator
        else:
            return InnerIterator(super().__iter__())

    def _worker_command(self, *cmd_args) -> List[Any]:
        """Executes a command in all workers and returns the results."""
        # print(f"cmd: {cmd_args}")
        for cmd_queue in self.cmd_queues:
            cmd_queue.put(cmd_args)
        # print(f"waiting for res")
        assert len(self.result_queues) == self.worker_config.num_workers
        res = {k: v for results_queue in self.result_queues for k, v in results_queue.get().items()}
        res = [res[i] for i in range(len(res))]
        # print(f"res: {res}")
        for r in res:
            if isinstance(r, Exception):
                raise r
        return res

    def _get_batch_size(self) -> Optional[int]:
        """Try to infer micro batch size from the dataset"""
        if isinstance(self.dataset, SavableDatasetWrapper):
            dataset = self.dataset.dataset
        else:
            dataset = self.dataset

        if (
            isinstance(dataset, BaseWrapperDataset)
            and (bds := dataset._find_wrapped_dataset(BatchDataset)) is not None
        ):
            assert isinstance(bds, BatchDataset)
            return bds.batch_size
        else:
            return None

    def save_state_rank(self) -> Optional[SavableDataLoaderState]:
        """
        Saves the state of the dataset for the current rank. Allows for restoring the state later
        using `restore_state_rank`, given the result of this method.

        Returns:
            The state of the dataset.
        """
        # Fetch current rank's worker's state
        if self.num_workers == 0:
            # No workers configured
            assert isinstance(self.dataset, SimpleSavableDatasetWrapper)
            worker_states = [self.dataset.save_state()]
            assert self._next_worker_id == 0
        elif self._persistent_iterator is None:
            # Workers configured, but not started yet.
            # If a state has already been restored, it will be returned.
            assert isinstance(self.dataset, SavableDatasetWrapper)
            worker_states = self.dataset.get_initial_checkpoint()
        else:
            # Fetch from worker processes
            worker_states = self._worker_command("get_checkpoint", self._worker_sample_counters)

        if worker_states is None:
            return None

        # Merge the states
        merged_state = SavableDataLoaderState(
            worker_states=worker_states,
            next_worker_id=self._next_worker_id,
            micro_batch_size=self._get_batch_size(),
        )

        # Not distributed -> return the merged state
        return merged_state

    def restore_state_rank(self, state: Optional[SavableDataLoaderState]) -> None:
        """
        Restores the saved state for the current rank.

        Args:
            state: The state to restore, as saved by `save_state_rank`.
        """
        assert self._persistent_iterator is None, "Cannot restore state while workers are running"
        if state is None:
            # Assume initial state
            return
        assert isinstance(state, SavableDataLoaderState)

        old_micro_batch_size = state.micro_batch_size
        micro_batch_size = self._get_batch_size()

        if isinstance(self.dataset, SavableDataset):
            assert micro_batch_size == old_micro_batch_size, (
                "Changing micro batch size is not allowed without workers"
            )

            assert len(state.worker_states) == 1
            assert isinstance(state.worker_states[0], FlexState)
            self.dataset.restore_state(state.worker_states[0])
        else:
            assert isinstance(self.dataset, SavableDatasetWrapper)
            assert all(isinstance(s, SavableDatasetCheckpoint) for s in state.worker_states)

            # Check batch sizes (before and after)
            if micro_batch_size != old_micro_batch_size:
                assert micro_batch_size is not None and old_micro_batch_size is not None, (
                    "Cannot resume with different batching mode "
                    "(batching to non-batching or vice versa)"
                )

                if micro_batch_size > old_micro_batch_size:
                    raise ValueError(
                        "Resuming with larger micro batch size is not allowed: "
                        f"{micro_batch_size} > {state.micro_batch_size}"
                    )
                elif (
                    micro_batch_size < old_micro_batch_size
                    and old_micro_batch_size % micro_batch_size != 0
                ):
                    raise ValueError(
                        "Resuming with smaller micro batch size only allowed if the old "
                        f"micro batch size is a multiple of the new one: {micro_batch_size} < {state.micro_batch_size}"
                    )
                batch_size_ratio = old_micro_batch_size // micro_batch_size
                for worker_state in state.worker_states:
                    assert isinstance(worker_state, SavableDatasetCheckpoint)
                    # When resuming with a smaller micro batch size, the offset must be scaled
                    # up to the new micro batch size to skip the same number of samples as before.
                    worker_state.offset *= batch_size_ratio

            self.dataset.restore_checkpoint(state.worker_states, worker_offset=state.next_worker_id)

            # Initialize the worker-sample counters so that every worker owns a valid
            # "last emitted sample" index.  Workers that have not emitted anything yet keep
            # the default value ``-1``.

            assert isinstance(state.worker_states, list)

            self._worker_sample_counters = [
                (
                    ws.state.sample_index - 1
                    if (isinstance(ws, SavableDatasetCheckpoint) and ws.state is not None)
                    else -1
                )
                for ws in state.worker_states
            ]

            self._next_worker_id = state.next_worker_id

    @deprecated(
        "`save_state` is deprecated and was renamed to `save_state_global` and will be removed "
        "in a future update. If you actually do not want to gather the states to a rank, use "
        "`save_state_rank` instead."
    )
    def save_state(self, dst_rank: int) -> Optional[Sequence[Optional[SavableDataLoaderState]]]:
        """Deprecated. Use `save_state_global` (or `save_state_rank`) instead."""

        return self.save_state_global(dst_rank)

    def save_state_global(
        self, global_dst_rank: int
    ) -> Optional[Sequence[Optional[SavableDataLoaderState]]]:
        """
        Saves the state of the dataset globally, collecting the state from all ranks using torch
        distributed. Allows for restoring the state later using `restore_state_global`, given the
        result of this method.
        Typical scenario: Save the state to disk only on the `dst_rank`, the other ranks do not
        save the state. Later, restore the state either only loaded on the `dst_rank` or
        loading on all ranks separately using `restore_state_global`.

        Note: If you want to save/restore the state per rank separately, use `save_state_rank` and
        the corresponding `restore_state_rank`. Also, these do not rely on torch distributed.

        Args:
            global_dst_rank: The state will be gathered to this rank. The rank refers to the
                global rank, not the rank within the data parallel group.

        Returns:
            The state of the dataset (or `None`, if not on `dst_rank`).
        """
        # Fetch current rank's worker's state
        merged_state = self.save_state_rank()

        # Gather the merged states
        if self.worker_config.world_size > 1:
            output: Optional[Sequence[Optional[SavableDataLoaderState]]]
            if self.worker_config.global_rank() == global_dst_rank:
                output = [None] * self.worker_config.world_size
            else:
                # Check if the global_dst_rank is in the same group at all
                if self.worker_config.data_parallel_group is not None:
                    try:
                        _ = torch.distributed.get_group_rank(
                            self.worker_config.data_parallel_group, global_dst_rank
                        )
                    except RuntimeError:
                        raise ValueError(
                            f"global_dst_rank {global_dst_rank} is not in the group of the current rank's worker config"
                        )

                output = None

            torch.distributed.gather_object(
                merged_state,
                output,
                global_dst_rank,
                group=self.worker_config.data_parallel_group,
            )

            return output
        else:
            # Not distributed -> return the merged state
            return [merged_state]

    @deprecated(
        "`restore_state` was renamed to `restore_state_global` and will be removed in a future update."
    )
    def restore_state(
        self,
        state: Optional[Sequence[Optional[SavableDataLoaderState]]],
    ) -> None:
        """Deprecated. Use `restore_state_global` (or `restore_state_rank`) instead."""

        return self.restore_state_global(state)

    def restore_state_global(
        self,
        state: Optional[Sequence[Optional[SavableDataLoaderState]]],
        *,
        src_rank: Optional[int] = None,
    ) -> None:
        """
        Restores the saved state from `save_state_global` (in torch distributed setup).
        The global state needs be loaded on every rank that has a data loader instance.

        Optionally, one can specify a src_rank and only provide the state once.
        In case of multiple data parallel groups, you must provide the state once
        in each data parallel group. In this case the `src_rank` is the rank within the
        data parallel group.

        Args:
            state: The state to restore, as saved by `save_state_global`.
            src_rank: The rank from which the state is broadcasted (within the data parallel group, if using DP groups).
        """

        assert self._persistent_iterator is None, "Cannot restore state while workers are running"

        # Only restore multi-rank if state is actually a list and we are in a distributed setup.
        # Otherwise treat as single rank state.
        if src_rank is None or self.worker_config.world_size == 1:
            assert isinstance(state, list), "State must be a list in distributed setup"
            assert len(state) == self.worker_config.world_size, (
                "State must be a list of size world_size"
            )

            # All ranks have the state
            # Select the state of the current rank
            rank_state = state[self.worker_config.rank]
        else:
            if self.worker_config.data_parallel_group is not None:
                # Only the src_rank has the state within this dp group
                try:
                    global_src_rank = torch.distributed.get_global_rank(
                        self.worker_config.data_parallel_group, src_rank
                    )
                except RuntimeError:
                    raise ValueError(
                        f"src_rank {src_rank} is not in the group of the current rank's worker config"
                    )
            else:
                # If no DP group is given, we assume the global rank is
                # the same as the data parallel rank
                global_src_rank = src_rank

            if self.worker_config.rank != src_rank:
                # Send the state to all other ranks
                assert state is None
                # Must still be a list of Nones
                state = [None] * self.worker_config.world_size
            else:
                assert isinstance(state, list), "State must be a list in distributed setup"
                assert len(state) == self.worker_config.world_size, (
                    "State must be a list of size world_size"
                )

            local_object = [None]
            torch.distributed.scatter_object_list(
                local_object,
                state,
                src=global_src_rank,
                group=self.worker_config.data_parallel_group,
            )
            rank_state = local_object[0]

        self.restore_state_rank(rank_state)

    def can_restore_sample(self) -> bool:
        return self.dataset.can_restore_sample()

    def restore_sample(self, restore_key: Tuple[Union[str, int, tuple], ...]) -> T:
        """Restores a sample from a key. This is useful to debug the dataset."""
        return self.dataset.restore_sample(restore_key)

    def config(self):
        """Get the configuration, which defines the dataset. Useful in conjunction with `save_state`
        and `restore_state` to match the configuration as well."""
        return {
            "type": type(self).__qualname__,
            "num_workers": self.num_workers,
            "persistent_workers": self.persistent_workers,
            "pin_memory": self.pin_memory,
            "prefetch_factor": None if self.num_workers == 0 else self.prefetch_factor,
            "dataset": self.dataset.config(),
        }


class BasicDataLoader(DataLoader[T], Generic[T]):
    """DataLoader that supports debugging the dataset without saving capability (e.g. for val/eval)."""

    #: The worker config
    worker_config: WorkerConfig
    #: The wrapped dataset. For multiprocessing, this is a :class:`megatron.energon.SavableDatasetWrapper`
    dataset: Union[SavableDatasetWrapper[T], SavableDataset[T]]

    id: int
    _sample_idx: int = 0

    def __init__(
        self,
        dataset: SavableDataset[T],
        gc_collect_every_n_steps: int = GC_DEFAULT_EVERY_N_ITER,
        gc_freeze_at_start: bool = True,
        prefetch_factor: int = 2,
        cache_pool: Optional[CachePool] = None,
        watchdog_timeout_seconds: Optional[float] = 60,
        watchdog_initial_timeout_seconds: Optional[float] = None,
        fail_on_timeout: bool = False,
    ):
        """
        Create the dataloader supporting saving and restoring the state.

        Args:
            dataset: The dataset to load.
            gc_collect_every_n_steps: The number of steps after which the garbage collector is
                called. As we're usually handling large (but few) tensors here, and the python
                garbage collection is already full of objects just by importing, this can improve
                the memory footprint quite a lot, and may even be necessary to avoid memory
                overflow.
            gc_freeze_at_start: If true, the garbage collector is frozen at the start of the worker
                processes. This improves the garbage collection performance by a lot.
                In rare cases, this may cause issues and can be disabled. Keep enabled if you
                experience no issues.
            cache_pool: If set, the cache pool to use for the dataset.
            watchdog_timeout_seconds: The timeout in seconds. If None, the watchdog is disabled.
            watchdog_initial_timeout_seconds: The initial timeout in seconds. If None, the timeout is the same as watchdog_timeout_seconds.
            fail_on_timeout: If True, stops the whole process upon timeout, after printing a stack trace.
        """
        self.worker_config = dataset.worker_config

        self.id = SavableDataLoader.next_id()

        dataset = WatchdogDataset(
            dataset,
            worker_config=self.worker_config,
            timeout_seconds=watchdog_timeout_seconds,
            initial_timeout_seconds=watchdog_initial_timeout_seconds,
            fail_on_timeout=fail_on_timeout,
        )

        if gc_collect_every_n_steps > 0:
            dataset = GcDataset(
                dataset,
                worker_config=self.worker_config,
                every_n_iter=gc_collect_every_n_steps,
                freeze=gc_freeze_at_start,
            )

        dataset = SimpleSavableDatasetWrapper(
            dataset, worker_config=self.worker_config, cache_pool=cache_pool
        )

        self._worker_sample_counters = [0] * max(self.worker_config.num_workers, 1)

        kwargs = {}
        if self.worker_config.num_workers > 0:
            # These must not be specified for num_workers =0
            kwargs["persistent_workers"] = True
            kwargs["prefetch_factor"] = prefetch_factor

        seed_per_worker = [
            self.worker_config.worker_seed(i) for i in range(self.worker_config.num_workers)
        ]

        gc.collect()  # This ensures that we don't include any old worker refs in the newly forked worker processes

        super().__init__(
            dataset,
            batch_size=None,
            shuffle=False,
            num_workers=self.worker_config.num_workers,
            pin_memory=True,
            worker_init_fn=partial(_init_worker, seed_per_worker),
            **kwargs,
        )
        if self.worker_config.should_log(level=1):
            self.worker_config.worker_log(
                {
                    "t": "BasicDataLoader.__init__",
                    "r": self.worker_config.rank,
                    "w": None,
                    "id": self.id,
                    "config": self.config(),
                }
            )

    def __iter__(self):
        outerself = self

        class InnerIterator:
            """Internal class which keeps the iterator alive across multiple `iter()` calls.
            If the inner iterator is exhausted, will also exhaust and a new instance is needed.
            Also saves the last sample index and the next worker id.
            """

            iter_idx: int = 0
            id: int

            def __init__(self, iterator):
                self._iterator = iterator
                self.id = SavableDataLoader.next_id()

                if outerself.worker_config.should_log(level=1):
                    outerself.worker_config.worker_log(
                        {
                            "t": "BasicDataLoader.iter",
                            "r": outerself.worker_config.rank,
                            "w": None,
                            "id": outerself.id,
                            "iter_id": self.id,
                        }
                    )

            def __iter__(self):
                return self

            def __next__(self):
                try:
                    worker_id, sample_idx, sample = next(self._iterator)
                    # If the next sample will be from the first worker, we can safely resume
                    self.next_worker_id = (worker_id + 1) % max(outerself.num_workers, 1)
                    if outerself.worker_config.should_log(level=1):
                        keys = default_get_keys(sample)
                        outerself.worker_config.worker_log(
                            {
                                **{
                                    "t": "BasicDataLoader.yield",
                                    "r": outerself.worker_config.rank,
                                    "w": None,
                                    "id": outerself.id,
                                    "iter_id": self.id,
                                    "worker_id": worker_id,
                                    "worker_idx": sample_idx,
                                    "idx": self.iter_idx,
                                    "iter_idx": self.iter_idx,
                                    "global_idx": outerself._sample_idx,
                                },
                                **({} if keys is None else {"keys": keys}),
                            }
                        )
                    outerself._sample_idx += 1
                    self.iter_idx += 1
                    return sample
                except StopIteration:
                    self.next_worker_id = 0
                    if outerself.worker_config.should_log(level=1):
                        outerself.worker_config.worker_log(
                            {
                                "t": "BasicDataLoader.StopIteration",
                                "r": outerself.worker_config.rank,
                                "w": None,
                                "id": outerself.id,
                                "iter_id": self.id,
                            }
                        )
                    raise

        return InnerIterator(super().__iter__())

    def config(self):
        """Get the configuration, which defines the dataset. Useful in conjunction with `save_state`
        and `restore_state` to match the configuration as well."""
        return {
            "type": type(self).__qualname__,
            "num_workers": self.num_workers,
            "persistent_workers": self.persistent_workers,
            "pin_memory": self.pin_memory,
            "prefetch_factor": None if self.num_workers == 0 else self.prefetch_factor,
            "dataset": self.dataset.config(),
        }

    def can_restore_sample(self) -> bool:
        return self.dataset.can_restore_sample()

    def restore_sample(self, restore_key: Tuple[Union[str, int, tuple], ...]) -> T:
        """Restores a sample from a key. This is useful to debug the dataset."""
        return self.dataset.restore_sample(restore_key)


def _sample_str(self, sample):
    """Returns a human readable debug string for a single sample, also uniquely identifying it."""
    import dataclasses
    import hashlib

    if isinstance(sample, torch.Tensor):
        return f"Tensor(shape={sample.shape}, dtype={sample.dtype}, sha256={hashlib.sha256(sample.detach().cpu().numpy().tobytes()).hexdigest()!r})"
    elif isinstance(sample, np.ndarray):
        return f"ndarray(shape={sample.shape}, dtype={sample.dtype}, sha256={hashlib.sha256(sample.tobytes()).hexdigest()!r})"
    elif isinstance(sample, (str, bytes)):
        if len(sample) > 100:
            return f"{sample[:100]!r}..."
        return repr(sample)
    elif isinstance(sample, (int, float)):
        return repr(sample)
    elif isinstance(sample, dict):
        innerstr = ", ".join(f"{k}={self._sample_str(v)}, " for k, v in sample.items())
        return f"dict({innerstr})"
    elif isinstance(sample, (list, tuple)):
        innerstr = ", ".join(self._sample_str(v) for v in sample)
        return f"{type(sample)}({innerstr})"
    elif dataclasses.is_dataclass(sample):
        innerstr = ", ".join(
            f"{field.name}={self._sample_str(getattr(sample, field.name))}, "
            for field in dataclasses.fields(sample)
        )
        return f"{type(sample).__name__}({innerstr})"
    else:
        return repr(sample)



# Copyright (c) 2025, NVIDIA CORPORATION.
# SPDX-License-Identifier: BSD-3-Clause

import json
from typing import List, Set

import braceexpand

from megatron.energon.edataclass import edataclass
from megatron.energon.epathlib import EPath
from megatron.energon.eyaml import load_yaml, load_yaml_json
from megatron.energon.flavors.webdataset.config import MAIN_FOLDER_NAME
from megatron.energon.flavors.webdataset.empty_dataset_error import EmptyDatasetError
from megatron.energon.flavors.webdataset.structs import (
    ShardInfo,
    WebdatasetInfo,
    WebdatasetSplits,
)
from megatron.energon.typed_converter import JsonParser


@edataclass
class WebdatasetMeta:
    """Class for getting metadata from a webdataset."""

    sample_excludes: Set[str]
    shards: List[ShardInfo]
    split_part_files: List[str]
    info_shard_files: List[str]

    @staticmethod
    def from_config(
        path: EPath,
        *,
        split_part: str,
        split_config: str = "split.yaml",
    ) -> "WebdatasetMeta":
        """
        Loads the metadata for a webdataset, i.e. the shards and sample excludes.

        Args:
            split_part: Which part to load (e.g. 'train', 'val', 'test').
            split_config: Config file to use for shard split definitions.
        """
        parser = JsonParser(strict=True)
        info_object = get_dataset_info(path)

        info = parser.raw_to_typed(
            info_object,
            WebdatasetInfo,
        )
        try:
            splits = parser.raw_to_typed(
                load_yaml_json(path / MAIN_FOLDER_NAME / split_config),
                WebdatasetSplits,
            )
        except FileNotFoundError:
            if split_config == "split.yaml":
                # Try split.json instead
                splits = parser.raw_to_typed(
                    load_yaml_json(path / MAIN_FOLDER_NAME / "split.json"),
                    WebdatasetSplits,
                )
            else:
                raise
        assert split_part in splits.split_parts, f"Invalid split part: {split_part!r}"
        split_excludes = {
            excluded
            for excluded in splits.exclude
            for excluded in braceexpand.braceexpand(excluded)
        }

        all_split_part_files = [
            name
            for name in splits.split_parts[split_part]
            for name in braceexpand.braceexpand(name)
        ]

        split_part_files = [name for name in all_split_part_files if name not in split_excludes]
        if len(split_part_files) == 0:
            raise EmptyDatasetError(f"No shards found in split part {split_part!r}")
        return WebdatasetMeta(
            sample_excludes={excluded for excluded in split_excludes if "/" in excluded},
            shards=[
                ShardInfo(
                    name=name,
                    path=path / name,
                    count=info.shard_counts[name],
                )
                for name in split_part_files
            ],
            split_part_files=all_split_part_files,
            info_shard_files=list(info.shard_counts.keys()),
        )


def get_info_shard_files(path: EPath) -> List[str]:
    """Use this if you don't need the full metadata for split parts, but just the shard files."""
    parser = JsonParser(strict=True)
    info = parser.raw_to_typed(
        get_dataset_info(path),
        WebdatasetInfo,
    )
    return list(info.shard_counts.keys())


def get_dataset_info(path: EPath) -> dict:
    """Given the path to an energon webdataset that contains a .nv-meta folder,
    return the dataset info as a dict.
    """

    info_config = path / MAIN_FOLDER_NAME / ".info.json"
    # YAML for backwards compatibility
    yaml_info_config = path / MAIN_FOLDER_NAME / ".info.yaml"

    if info_config.is_file():
        with info_config.open("r") as rf:
            return json.load(rf)
    elif yaml_info_config.is_file():
        return load_yaml(yaml_info_config.read_bytes())
    else:
        raise ValueError(f"No info config file found at {info_config} or {yaml_info_config}")


def check_dataset_info_present(path: EPath) -> bool:
    """Given the path to an energon webdataset that contains a .nv-meta folder,
    return True if the dataset info is present, False otherwise.
    """
    return (path / MAIN_FOLDER_NAME / ".info.json").is_file() or (
        path / MAIN_FOLDER_NAME / ".info.yaml"
    ).is_file()



# Copyright (c) 2025, NVIDIA CORPORATION.
# SPDX-License-Identifier: BSD-3-Clause
"""Provides methods for converting typed objects to json objects and vice versa."""

import dataclasses
import importlib
import inspect
import json
import re
import typing
from typing import Any, Callable, Dict, Literal, Optional, Tuple, Type, TypeVar, Union

from megatron.energon.edataclass import edataclass
from megatron.energon.epathlib import EPath


class JsonValueError(ValueError):
    """Internal error class, for better error messages."""

    def __init__(
        self, msg: str, expected_type: Type, value: Any, path: str, stage: Tuple[int, ...]
    ) -> None:
        """New error"""
        super().__init__(msg)
        self.expected_type = expected_type
        self.value = value
        self.path = path
        self.stage = stage


class _MissingValue:
    """Internal class representing a missing value (i.e. a field which has not been specified,
    opposed to explicit `None`) for better error messages."""

    def __repr__(self) -> str:
        return "missing value"


_missing_value = _MissingValue()

TType = TypeVar("TType")


def _check_instance_type(cls, inst_type: Type) -> bool:
    """Check if a class is an instance of a type."""
    if inst_type is None:
        return True
    elif typing.get_origin(inst_type) is not None:
        org = typing.get_origin(inst_type)
        if org is Union:
            for check_type in typing.get_args(inst_type):
                if _check_instance_type(cls, check_type):
                    return True
        elif isinstance(org, type) and issubclass(cls, org):
            return True
        return False
    elif inst_type is Any:
        return True
    else:
        return not issubclass(inst_type, type) or issubclass(cls, inst_type)


@edataclass
class JsonParser:
    strict: bool = True

    def _resolve_object(
        self,
        module_name: str,
        object_name: str,
        cls: Type[TType],
        is_type: bool,
        is_callable: bool,
        is_instantiating_class: bool,
        is_calling_function: bool,
    ) -> TType:
        """Resolve an object reference to the object."""
        try:
            module = importlib.import_module(module_name)
        except ModuleNotFoundError:
            try_energon_module = importlib.import_module("megatron.energon", package=None)
            if hasattr(try_energon_module, object_name):
                module = try_energon_module
            else:
                raise
        try:
            return getattr(module, object_name)
        except AttributeError:
            raise ModuleNotFoundError(f"Object {object_name} not found in {module_name}")

    def raw_to_instance(
        self,
        kwargs: dict,
        inst_type: Type[TType],
        _path: str = "root",
        _stage: Tuple[int, ...] = (),
    ) -> TType:
        """
        Try to import and instantiate a class from a dict with "__module__" and "__class__"/"__function__" keys.

        Args:
            kwargs: The dict to parse
            inst_type: Expected return type, used if type is not specified in the kwargs
            strict: If true, don't allow additional attributes
            _path: (internal for recursive call) The path to the object being converted from the root
            _stage: (internal for recursive call) Numbers representing the position of the current
                object being converted from the root

        Returns:
            Instantiated class
        """
        kwargs = kwargs.copy()
        module_name = kwargs.pop("__module__", None)
        # Check if this is a type of Type[...] or just a class. Type[...] will return the class instead
        # of instantiating it.
        is_type = typing.get_origin(inst_type) is type
        is_callable = typing.get_origin(inst_type) is typing.get_origin(Callable)
        is_calling_function = False
        is_instantiating_class = False
        if is_type:
            inst_type = typing.get_args(inst_type)[0]
            object_name = kwargs.pop("__class__", None)
            if module_name is None or object_name is None:
                raise JsonValueError(
                    f"Expected __module__ and __class__ for Type[{inst_type}], got {kwargs}",
                    inst_type,
                    (module_name, object_name),
                    _path,
                    _stage,
                )
        elif is_callable:
            object_name = kwargs.pop("__function__", None)
            if module_name is None or object_name is None:
                raise JsonValueError(
                    f"Expected __module__ and __function__ for {inst_type}, got {kwargs}",
                    inst_type,
                    (module_name, object_name),
                    _path,
                    _stage,
                )
        else:
            if "__class__" in kwargs:
                object_name = kwargs.pop("__class__", None)
                is_instantiating_class = True
                is_calling_function = False
            elif "__function__" in kwargs:
                object_name = kwargs.pop("__function__", None)
                is_instantiating_class = False
                is_calling_function = True
            # Else case: It's a plain type, and nothing was passed, use the default cls
        if module_name is None or object_name is None:
            cls = inst_type
        else:
            cls = self._resolve_object(
                module_name,
                object_name,
                inst_type,
                is_type,
                is_callable,
                is_instantiating_class,
                is_calling_function,
            )

            if is_type:
                if isinstance(inst_type, type) and (
                    not isinstance(cls, type) or not issubclass(cls, inst_type)
                ):
                    raise JsonValueError(
                        f"Expected Type[{inst_type}], got {cls}", inst_type, cls, _path, _stage
                    )
            elif is_callable:
                if not callable(cls):
                    raise JsonValueError(
                        f"Expected a callable, got {cls}", inst_type, cls, _path, _stage
                    )
            elif is_instantiating_class:
                if not isinstance(cls, type) or not _check_instance_type(cls, inst_type):
                    raise JsonValueError(
                        f"Expected {inst_type}, got {cls}", inst_type, cls, _path, _stage
                    )
            else:
                assert is_calling_function
                if not callable(cls):
                    raise JsonValueError(
                        f"Expected {inst_type}, got {cls}", inst_type, cls, _path, _stage
                    )
        if is_type or is_callable:
            inst = cls
        else:
            # Do not assert the other cases, we fallback to the passed cls
            inst = self.safe_call_function(kwargs, cls, allow_imports=True)
            assert not isinstance(cls, type) or _check_instance_type(type(inst), inst_type), (
                f"Expected {inst_type}, got {cls}"
            )
        return inst

    def raw_to_typed(  # noqa: C901
        self,
        raw_data: Union[dict, list, str, int, bool, float, None],
        inst_type: Type[TType],
        allow_imports: bool = False,
        _path: str = "root",
        _stage: Tuple[int, ...] = (),
    ) -> TType:
        """
        Converts raw data (i.e. dicts, lists and primitives) to typed objects (like
        `NamedTuple` or `dataclasses.dataclass`). Validates that python typing matches.

        Usage::

            class MyNamedTuple(NamedTuple):
                x: int
                y: str

            assert raw_to_typed({'x': int, 'y': "foo"}, MyNamedTuple) == MyNamedTuple(x=5, y="foo")

        Args:
            raw_data: The raw (e.g. json) data to be made as `inst_type`
            inst_type: The type to return
            allow_imports: If true, parse '__module__' and '__class__/__function__' attributes to allow explicit
                instantiation of types
            _path: (internal for recursive call) The path to the object being converted from the root
            _stage: (internal for recursive call) Numbers representing the position of the current
                object being converted from the root

        Returns:
            The input data as `inst_type`.
        """
        type_name = getattr(inst_type, "__name__", repr(inst_type))
        if raw_data is _missing_value:
            raise JsonValueError(
                f"Missing value at {_path}",
                inst_type,
                raw_data,
                _path,
                _stage,
            )
        elif inst_type in (str, int, float, bool, None, type(None)):
            # Literal types or missing data
            if not isinstance(raw_data, inst_type) and not (
                isinstance(raw_data, int) and inst_type is float
            ):
                raise JsonValueError(
                    f"Type does not match, expected {type_name} at {_path}, got {raw_data!r}",
                    inst_type,
                    raw_data,
                    _path,
                    _stage,
                )
            return raw_data
        elif inst_type is Any:
            if (
                allow_imports
                and isinstance(raw_data, dict)
                and "__module__" in raw_data
                and ("__class__" in raw_data or "__function__" in raw_data)
            ):
                return self.raw_to_instance(raw_data, inst_type, _path=_path, _stage=_stage)
            # Any
            return raw_data
        elif typing.get_origin(inst_type) is Literal:
            # Literal[value[, ...]]
            values = typing.get_args(inst_type)
            if raw_data not in values:
                raise JsonValueError(
                    f"Expected {type_name} at {_path}, got {raw_data!r}",
                    inst_type,
                    raw_data,
                    _path,
                    _stage,
                )
            return raw_data
        elif typing.get_origin(inst_type) is Union:
            # Union[union_types[0], union_types[1], ...]
            union_types = typing.get_args(inst_type)
            if None in union_types:
                # Fast Optional path
                if raw_data is None:
                    return None
            best_inner_error: Optional[JsonValueError] = None
            inner_exceptions = []
            for subtype in union_types:
                try:
                    return self.raw_to_typed(
                        raw_data,
                        subtype,
                        allow_imports,
                        f"{_path} -> {getattr(subtype, '__name__', repr(subtype))}",
                        _stage + (1,),
                    )
                except JsonValueError as err:
                    if best_inner_error is None or len(err.stage) > len(best_inner_error.stage):
                        best_inner_error = err
                        inner_exceptions.clear()
                        inner_exceptions.append(err)
                    elif len(err.stage) == len(best_inner_error.stage):
                        inner_exceptions.append(err)
                    continue
            if len(inner_exceptions) > 0:
                cur_exc = inner_exceptions[0]
                for next_exc in inner_exceptions[1:]:
                    try:
                        raise next_exc from cur_exc
                    except JsonValueError as e:
                        cur_exc = e
                raise cur_exc
            else:
                raise JsonValueError(
                    f"Expected {inst_type} at {_path}, got {raw_data!r}",
                    inst_type,
                    raw_data,
                    _path,
                    _stage,
                )
        elif (
            isinstance(inst_type, type)
            and issubclass(inst_type, tuple)
            and hasattr(inst_type, "__annotations__")
        ):
            # class MyClass(NamedTuple): ...
            if not isinstance(raw_data, dict):
                raise JsonValueError(
                    f"Expected {type_name} at {_path}, got {raw_data!r}",
                    inst_type,
                    raw_data,
                    _path,
                    _stage,
                )
            if getattr(inst_type, "__dash_keys__", "False"):
                raw_data = {key.replace("-", "_"): val for key, val in raw_data.items()}
            defaults = getattr(inst_type, "_field_defaults", {})
            kwargs = {
                field_name: self.raw_to_typed(
                    raw_data.get(field_name, defaults.get(field_name, _missing_value)),
                    field_type,
                    allow_imports,
                    f"{_path} -> {type_name}:{field_name}",
                    _stage + (idx,),
                )
                for idx, (field_name, field_type) in enumerate(inst_type.__annotations__.items())
            }
            if self.strict and not set(raw_data).issubset(inst_type.__annotations__):
                raise JsonValueError(
                    f"Additional attributes for {type_name} at {_path}, got {raw_data!r}",
                    inst_type,
                    raw_data,
                    _path,
                    _stage,
                )
            try:
                return inst_type(**kwargs)
            except BaseException:
                raise JsonValueError(
                    f"Expected {type_name} at {_path}, got {raw_data!r}",
                    inst_type,
                    raw_data,
                    _path,
                    _stage,
                )
        elif dataclasses.is_dataclass(inst_type):
            # dataclass
            if not isinstance(raw_data, dict):
                raise JsonValueError(
                    f"Expected {type_name} at {_path}, got {raw_data!r}",
                    inst_type,
                    raw_data,
                    _path,
                    _stage,
                )
            kwargs = {
                field.name: self.raw_to_typed(
                    raw_data.get(
                        field.name,
                        (
                            (
                                _missing_value
                                if field.default_factory is dataclasses.MISSING
                                else field.default_factory()
                            )
                            if field.default is dataclasses.MISSING
                            else field.default
                        ),
                    ),
                    field.type,
                    allow_imports,
                    f"{_path} -> {type_name}:{field.name}",
                    _stage + (idx,),
                )
                for idx, field in enumerate(dataclasses.fields(inst_type))
                if field.init
            }
            if self.strict and not set(raw_data).issubset(
                field.name for field in dataclasses.fields(inst_type) if field.init
            ):
                raise JsonValueError(
                    f"Additional attributes for {type_name} at {_path}, got {raw_data!r}",
                    inst_type,
                    raw_data,
                    _path,
                    _stage,
                )
            try:
                return inst_type(**kwargs)
            except BaseException:
                raise JsonValueError(
                    f"Expected {type_name} at {_path}, got {raw_data!r}",
                    inst_type,
                    raw_data,
                    _path,
                    _stage,
                )
        elif typing.get_origin(inst_type) is list:
            # List[inner_type]
            (inner_type,) = typing.get_args(inst_type)
            if not isinstance(raw_data, list):
                raise JsonValueError(
                    f"Expected {type_name} at {_path}, got {raw_data!r}",
                    inst_type,
                    raw_data,
                    _path,
                    _stage,
                )
            return [
                self.raw_to_typed(
                    val, inner_type, allow_imports, f"{_path} -> {idx}", _stage + (idx,)
                )
                for idx, val in enumerate(raw_data)
            ]
        elif typing.get_origin(inst_type) is set:
            # Set[inner_type]
            (inner_type,) = typing.get_args(inst_type)
            if not isinstance(raw_data, list):
                raise JsonValueError(
                    f"Expected {type_name} at {_path}, got {raw_data!r}",
                    inst_type,
                    raw_data,
                    _path,
                    _stage,
                )
            res = set(
                self.raw_to_typed(
                    val, inner_type, allow_imports, f"{_path} -> {idx}", _stage + (idx,)
                )
                for idx, val in enumerate(raw_data)
            )
            if len(res) != len(raw_data):
                raise JsonValueError(
                    f"Duplicate element at {_path}",
                    inst_type,
                    raw_data,
                    _path,
                    _stage,
                )
            return res
        elif typing.get_origin(inst_type) is tuple:
            # Tuple[inner_types[0], inner_types[1], ...] or Tuple[inner_types[0], Ellipsis/...]
            inner_types = typing.get_args(inst_type)
            if not isinstance(raw_data, list):
                raise JsonValueError(
                    f"Expected {type_name} at {_path}, got {raw_data!r}",
                    inst_type,
                    raw_data,
                    _path,
                    _stage,
                )
            if len(inner_types) == 2 and inner_types[1] is Ellipsis:
                # Tuple of arbitrary length, all elements same type
                # Tuple[inner_types[0], Ellipsis/...]
                return tuple(
                    self.raw_to_typed(
                        val, inner_types[0], allow_imports, f"{_path} -> {idx}", _stage + (idx,)
                    )
                    for idx, val in enumerate(raw_data)
                )
            else:
                # Fixed size/typed tuple
                # Tuple[inner_types[0], inner_types[1], ...]
                if len(raw_data) != len(inner_types):
                    raise JsonValueError(
                        f"Expected {type_name} at {_path}, got {raw_data!r}",
                        inst_type,
                        raw_data,
                        _path,
                        _stage,
                    )
                return [
                    self.raw_to_typed(
                        val, inner_type, allow_imports, f"{_path} -> {idx}", _stage + (idx,)
                    )
                    for idx, (val, inner_type) in enumerate(zip(raw_data, inner_types))
                ]
        elif typing.get_origin(inst_type) is dict:
            # Dict[str, value_type]
            key_type, value_type = typing.get_args(inst_type)
            assert key_type is str
            if not isinstance(raw_data, dict):
                raise JsonValueError(
                    f"Expected {type_name} at {_path}, got {raw_data!r}",
                    inst_type,
                    raw_data,
                    _path,
                    _stage,
                )
            return {
                key: self.raw_to_typed(
                    val, value_type, allow_imports, f"{_path} -> {key!r}", _stage + (idx,)
                )
                for idx, (key, val) in enumerate(raw_data.items())
            }
        elif inst_type in (dict, list):
            # dict, list (no subtyping)
            if not isinstance(raw_data, inst_type):
                raise JsonValueError(
                    f"Expected {type_name} at {_path}, got {raw_data!r}",
                    inst_type,
                    raw_data,
                    _path,
                    _stage,
                )
            return raw_data
        elif inst_type is EPath:
            if isinstance(raw_data, str):
                return EPath(raw_data)
            elif not isinstance(raw_data, EPath):
                raise JsonValueError(
                    f"Expected {type_name} at {_path}, got {raw_data!r}",
                    inst_type,
                    raw_data,
                    _path,
                    _stage,
                )
            return raw_data
        elif (
            allow_imports
            and isinstance(raw_data, dict)
            and "__module__" in raw_data
            and ("__class__" in raw_data or "__function__" in raw_data)
        ):
            return self.raw_to_instance(raw_data, inst_type, _path=_path, _stage=_stage)
        else:
            return raw_data

    def safe_call_function(
        self,
        raw_data: Union[dict, list, str, int, bool, float, None],
        fn: Callable[..., TType],
        allow_imports: bool = False,
    ) -> TType:
        """
        Converts raw data (i.e. dicts, lists and primitives) to typed call arguments.
        Validates that python typing matches.

        Usage::

            def fn(arg1: float, arg2: MyType, arg3) -> Any:
                assert isinstance(arg1, float)
                assert isinstance(arg2, MyType)

            fn(3.141, MyType(), None)

        Args:
            raw_data: The raw (e.g. json) data to be made as `inst_type`
            fn: The function to call with the converted data
            strict: If true, don't allow additional attributes
            allow_imports: If true, allow instantiating objects by specifying __module__ and __class__/__function__.

        Returns:
            The return value of `fn`
        """
        parameters = list(inspect.signature(fn).parameters.items())
        if inspect.isclass(fn):
            init_sig = getattr(fn, "__init__", None)
            if init_sig is not None:
                parameters = list(inspect.signature(init_sig).parameters.items())[1:]
        args = []
        kwargs = {}
        if isinstance(raw_data, dict):
            unused_args = raw_data.copy()
            for idx, (key, param) in enumerate(parameters):
                t = Any if param.annotation is inspect.Parameter.empty else param.annotation
                if param.kind in (
                    inspect.Parameter.POSITIONAL_OR_KEYWORD,
                    inspect.Parameter.KEYWORD_ONLY,
                ):
                    if param.default is inspect.Parameter.empty and key not in unused_args:
                        raise ValueError(f"Missing required argument {key!r} for {fn}")
                    kwargs[key] = self.raw_to_typed(
                        unused_args.pop(key, param.default),
                        t,
                        allow_imports,
                        _path=key,
                        _stage=(idx,),
                    )
                elif param.kind == inspect.Parameter.VAR_KEYWORD:
                    for arg_key, arg_val in unused_args.items():
                        kwargs[arg_key] = self.raw_to_typed(
                            arg_val, t, allow_imports, _path=key, _stage=(idx,)
                        )
                    unused_args.clear()
                elif param.kind == inspect.Parameter.VAR_POSITIONAL:
                    # No way to pass positional arguments
                    pass
                elif param.kind == inspect.Parameter.POSITIONAL_ONLY:
                    # No way to pass positional arguments
                    raise RuntimeError(f"Unsupported positional only argument {key!r}")
                else:
                    raise RuntimeError(f"Unknown parameter kind {param.kind!r}")
            if self.strict and len(unused_args) > 0:
                raise ValueError(f"Unexpected arguments: {unused_args!r}")
        elif isinstance(raw_data, list):
            unused_args = raw_data.copy()
            for idx, (key, param) in enumerate(parameters):
                t = Any if param.annotation is inspect.Parameter.empty else param.annotation
                if param.kind == inspect.Parameter.POSITIONAL_ONLY:
                    if param.default is inspect.Parameter.empty and len(unused_args) == 0:
                        raise ValueError(
                            f"Missing required positional-only argument {key!r} at index {idx}"
                        )
                    args.append(
                        self.raw_to_typed(
                            unused_args.pop(), t, allow_imports, _path=key, _stage=(idx,)
                        )
                    )
                elif param.kind == inspect.Parameter.POSITIONAL_OR_KEYWORD:
                    if param.default is inspect.Parameter.empty and len(unused_args) == 0:
                        raise ValueError(
                            f"Missing required positional argument {key!r} at index {idx}"
                        )
                    if len(unused_args) == 0:
                        arg_val = param.default
                    else:
                        arg_val = unused_args.pop()
                    args.append(
                        self.raw_to_typed(arg_val, t, allow_imports, _path=key, _stage=(idx,))
                    )
                elif param.kind == inspect.Parameter.VAR_POSITIONAL:
                    for arg_val in unused_args:
                        args.append(
                            self.raw_to_typed(arg_val, t, allow_imports, _path=key, _stage=(idx,))
                        )
                    unused_args.clear()
                elif param.kind == inspect.Parameter.VAR_KEYWORD:
                    # No way to pass keyword arguments
                    pass
                elif param.kind == inspect.Parameter.KEYWORD_ONLY:
                    raise RuntimeError(f"Unsupported keyword-only argument {key!r}")
                else:
                    raise RuntimeError(f"Unknown parameter kind {param.kind!r}")
            if self.strict and len(unused_args) > 0:
                raise ValueError(f"Unexpected arguments: {unused_args!r}")
        else:
            raise ValueError(
                f"Cannot call function with raw data of type {type(raw_data)!r}, require list or dict"
            )
        return fn(*args, **kwargs)

    def override(  # noqa: C901
        self,
        value: TType,
        overrides: Any,
        inst_type: Optional[Type[TType]] = None,
        allow_imports: bool = False,
        _path: str = "root",
        _stage: Tuple[int, ...] = (),
    ) -> TType:
        """
        Allows overriding values of a typed object using environment config.
        Allows overriding single config variables, or whole objects.

        Examples::

            class MyNamedTuple(NamedTuple):
                x: int
                y: str

            class MyNested(NamedTuple):
                nested: MyNamedTuple

            assert override(
                MyNested(nested=MyNamedTuple(x=42, y="foo")),
                {'nested.x': 5},
            ) == MyNested(nested=MyNamedTuple(x=5, y="foo"))
            assert override(
                MyNested(nested=MyNamedTuple(x=42, y="foo")),
                {'nested': '{"x": 5, "y": "bar"}'},
            ) == MyNested(nested=MyNamedTuple(x=5, y="bar"))

        Args:
            value: The base value to override.
            overrides: The overrides to apply
            strict: If true, no additional keys are allowed
            inst_type: If given, validate against this base type instead of the type of `value`.
            allow_imports: If true, allow instantiating types with dicts of __module__ and __class__/__function__.
            _path: Internal: The path to the current value.
            _stage: Internal: The current stage of the override.

        Returns:
            Same type as the input object (or `inst_type` if set), copied and updated from the
            overrides.
        """
        if inst_type is None:
            inst_type = type(value)
        type_name = getattr(inst_type, "__name__", repr(inst_type))
        if inst_type in (str, int, float, bool, None, type(None)):
            # Literal types
            if inst_type in (None, type(None)) and overrides == "None":
                overrides = None
            elif inst_type is bool and overrides in ("True", "true", "1", "False", "false", "0"):
                overrides = overrides in ("True", "true", "1")
            elif inst_type in (int, float) and isinstance(overrides, str):
                overrides = inst_type(overrides)
            if not isinstance(overrides, inst_type) and not (
                isinstance(overrides, int) and inst_type is float
            ):
                raise JsonValueError(
                    f"Type does not match, expected {type_name} at {_path}, got {overrides!r}",
                    inst_type,
                    overrides,
                    _path,
                    _stage,
                )
            return overrides
        elif inst_type is Any:
            # Any
            if isinstance(overrides, str):
                if overrides.isnumeric():
                    return int(overrides)
                elif overrides == "True":
                    return True
                elif overrides == "False":
                    return True
                return overrides
            if isinstance(value, (dict, list, tuple)):
                # Merge with dict, list, str
                return self.override(value, overrides, type(value), allow_imports, _path, _stage)
            raise JsonValueError(
                f"Expected {type_name} at {_path}, got {overrides!r}",
                inst_type,
                overrides,
                _path,
                _stage,
            )
        elif typing.get_origin(inst_type) is Literal:
            # Literal[value]
            (value,) = typing.get_args(inst_type)
            if value != overrides:
                raise JsonValueError(
                    f"Expected {type_name} at {_path}, got {overrides!r}",
                    inst_type,
                    overrides,
                    _path,
                    _stage,
                )
            return value
        elif typing.get_origin(inst_type) is Union:
            # Union[union_types[0], union_types[1], ...]
            union_types = typing.get_args(inst_type)
            if isinstance(overrides, str):
                for subtype in union_types:
                    if subtype is None and overrides == "None":
                        return None
                    elif subtype is bool:
                        if overrides == "True":
                            return True
                        elif overrides == "False":
                            return False
                    elif subtype is int and overrides.strip().isnumeric():
                        return int(overrides)
                    elif subtype is str:
                        return overrides
                    elif subtype is float and float_pattern.fullmatch(overrides):
                        return float(overrides)
                if overrides.lstrip().startswith("{") or overrides.lstrip().startswith("["):
                    overrides = json.loads(overrides)
                return self.raw_to_typed(
                    overrides,
                    inst_type,
                    allow_imports,
                    _path,
                    _stage,
                )
            for subtype in union_types:
                if _isinstance_deep(value, subtype):
                    return self.override(
                        value,
                        overrides,
                        subtype,
                        allow_imports,
                        f"{_path} -> {getattr(subtype, '__name__', repr(subtype))}",
                        _stage + (1,),
                    )
            raise JsonValueError(
                f"Expected {type_name} at {_path}, existing is {value!r} which is invalid",
                inst_type,
                value,
                _path,
                _stage,
            )
        elif (
            isinstance(inst_type, type)
            and issubclass(inst_type, tuple)
            and hasattr(inst_type, "__annotations__")
        ):
            # class MyClass(NamedTuple): ...
            if not isinstance(overrides, (dict, str)):
                raise JsonValueError(
                    f"Expected {type_name} at {_path}, got {overrides!r}",
                    inst_type,
                    overrides,
                    _path,
                    _stage,
                )
            if isinstance(overrides, str):
                return self.raw_to_typed(
                    json.loads(overrides),
                    inst_type,
                    allow_imports,
                    _path,
                    _stage,
                )
            local_overrides = _split_dict_keys(overrides)
            if getattr(inst_type, "__dash_keys__", "False"):
                local_overrides = {
                    key.replace("-", "_"): val for key, val in local_overrides.items()
                }
            kwargs = {
                field_name: (
                    self.override(
                        getattr(value, field_name),
                        local_overrides.pop(field_name),
                        field_type,
                        allow_imports,
                        f"{_path} -> {type_name}:{field_name}",
                        _stage + (idx,),
                    )
                    if field_name in local_overrides
                    else getattr(value, field_name)
                )
                for idx, (field_name, field_type) in enumerate(inst_type.__annotations__.items())
            }
            if self.strict and len(local_overrides) != 0:
                raise JsonValueError(
                    f"Invalid config keys {', '.join(local_overrides.keys())} for {type_name} at "
                    f"{_path}",
                    inst_type,
                    overrides,
                    _path,
                    _stage,
                )
            try:
                return inst_type(**kwargs)
            except BaseException:
                raise JsonValueError(
                    f"Expected {type_name} at {_path}, got {overrides!r}",
                    inst_type,
                    overrides,
                    _path,
                    _stage,
                )
        elif dataclasses.is_dataclass(inst_type):
            # dataclass
            if not isinstance(overrides, (dict, str)):
                raise JsonValueError(
                    f"Expected {type_name} at {_path}, got {overrides!r}",
                    inst_type,
                    overrides,
                    _path,
                    _stage,
                )
            if isinstance(overrides, str):
                return self.raw_to_typed(
                    json.loads(overrides),
                    inst_type,
                    allow_imports,
                    _path,
                    _stage,
                )
            local_overrides = _split_dict_keys(overrides)
            if getattr(inst_type, "__dash_keys__", "False"):
                local_overrides = {
                    key.replace("-", "_"): val for key, val in local_overrides.items()
                }
            kwargs = {
                field.name: (
                    self.override(
                        getattr(value, field.name),
                        local_overrides.pop(field.name),
                        field.type,
                        allow_imports,
                        f"{_path} -> {type_name}:{field.name}",
                        _stage + (idx,),
                    )
                    if field.name in local_overrides
                    else getattr(value, field.name)
                )
                for idx, field in enumerate(dataclasses.fields(inst_type))
                if field.init
            }
            if self.strict and len(local_overrides) != 0:
                raise JsonValueError(
                    f"Invalid config keys {', '.join(local_overrides.keys())} for {type_name} at "
                    f"{_path}",
                    inst_type,
                    overrides,
                    _path,
                    _stage,
                )
            try:
                return inst_type(**kwargs)
            except BaseException:
                raise JsonValueError(
                    f"Expected {type_name} at {_path}, got {overrides!r}",
                    inst_type,
                    overrides,
                    _path,
                    _stage,
                )
        elif (
            typing.get_origin(inst_type) is list
            or typing.get_origin(inst_type) is tuple
            or inst_type in (list, tuple)
        ):
            # List[inner_type] or Tuple[inner_type, Ellipsis] or
            # Tuple[inner_type[0], inner_type[1], ...]
            if inst_type is list:
                inner_type = Any
                inner_types = []
                cls = list
            elif inst_type is tuple:
                inner_type = Any
                inner_types = []
                cls = tuple
            elif typing.get_origin(inst_type) is list:
                (inner_type,) = typing.get_args(inst_type)
                inner_types = []
                cls = list
            else:
                inner_types = typing.get_args(inst_type)
                if len(inner_types) == 2 and inner_types[1] is Ellipsis:
                    inner_type = inner_types[0]
                else:
                    inner_type = None
                cls = tuple
            if not isinstance(overrides, (dict, str)):
                raise JsonValueError(
                    f"Expected {type_name} at {_path}, got {overrides!r}",
                    inst_type,
                    overrides,
                    _path,
                    _stage,
                )
            if isinstance(overrides, str):
                return self.raw_to_typed(
                    json.loads(overrides),
                    inst_type,
                    allow_imports,
                    _path,
                    _stage,
                )
            local_overrides = _split_dict_keys(overrides)
            if not all(key.isnumeric() for key in local_overrides.keys()):
                raise JsonValueError(
                    f"Expected {type_name} at {_path}, got {overrides!r}, expected integer keys",
                    inst_type,
                    overrides,
                    _path,
                    _stage,
                )
            local_overrides_int = {int(key): value for key, value in local_overrides.items()}
            new_max_idx = max(local_overrides_int.keys())
            original_max_idx = len(value)
            if inner_type is None and new_max_idx >= len(inner_types):
                raise JsonValueError(
                    f"Expected {type_name} at {_path}, got {overrides!r}, index {new_max_idx} out of "
                    f"bounds",
                    inst_type,
                    overrides,
                    _path,
                    _stage,
                )
            for i in range(original_max_idx, new_max_idx):
                if i not in local_overrides_int:
                    raise JsonValueError(
                        f"Expected {type_name} at {_path}, got {overrides!r}, missing value for index "
                        f"{i}",
                        inst_type,
                        overrides,
                        _path,
                        _stage,
                    )
            return cls(
                (
                    self.override(
                        value[idx],
                        local_overrides_int[idx],
                        inner_type,
                        allow_imports,
                        f"{_path} -> {idx}",
                        _stage + (idx,),
                    )
                    if idx in local_overrides_int
                    else value[idx]
                )
                for idx in range(max(new_max_idx + 1, original_max_idx))
            )
        elif typing.get_origin(inst_type) is dict or inst_type is dict:
            # Dict[str, value_type]
            if inst_type is dict:
                value_type = Any
            else:
                key_type, value_type = typing.get_args(inst_type)
                assert key_type is str
            if not isinstance(overrides, (dict, str)):
                raise JsonValueError(
                    f"Expected {type_name} at {_path}, got {overrides!r}",
                    inst_type,
                    overrides,
                    _path,
                    _stage,
                )
            if isinstance(overrides, str):
                return self.raw_to_typed(
                    json.loads(overrides),
                    inst_type,
                    allow_imports,
                    _path,
                    _stage,
                )
            local_overrides = _split_dict_keys(overrides)
            if getattr(inst_type, "__dash_keys__", "False"):
                local_overrides = {
                    key.replace("-", "_"): val for key, val in local_overrides.items()
                }
            res = {
                key: (
                    self.override(
                        subvalue,
                        local_overrides.pop(key),
                        value_type,
                        allow_imports,
                        f"{_path} -> {type_name}:{key!r}",
                        _stage + (idx,),
                    )
                    if key in local_overrides
                    else subvalue
                )
                for idx, (key, subvalue) in value.items()
            }
            for key, val in local_overrides.items():
                if not isinstance(val, str):
                    raise JsonValueError(
                        f"Expected new {type_name} at {_path} -> {type_name}:{key!r}, got {val!r}",
                        inst_type,
                        overrides,
                        _path,
                        _stage,
                    )
                res[key] = self.raw_to_typed(
                    json.loads(val),
                    value_type,
                    allow_imports,
                    f"{_path} -> {type_name}:{key!r}",
                    _stage + (len(res),),
                )
            return res
        else:
            raise RuntimeError(f"Unknown type {inst_type}")


def to_json_object(obj: Any) -> Any:
    """
    Converts the given object to a json object.

    Args:
        obj: The object to convert

    Returns:
        The json-like object.
    """
    if isinstance(obj, (str, int, float, bool, type(None))):
        # Literal types
        return obj
    elif isinstance(obj, tuple) and hasattr(obj, "__annotations__"):
        # class MyClass(NamedTuple): ...
        return {
            field_name: to_json_object(getattr(obj, field_name))
            for field_name in obj.__annotations__.keys()
        }
    elif dataclasses.is_dataclass(obj):
        # dataclass
        return {
            field.name: to_json_object(getattr(obj, field.name))
            for field in dataclasses.fields(obj)
            if field.init
        }
    elif isinstance(obj, (list, tuple)):
        return [to_json_object(val) for val in obj]
    elif isinstance(obj, dict):
        return {key: to_json_object(val) for key, val in obj.items()}
    else:
        raise RuntimeError(f"Unknown type {type(obj)}")


float_pattern = re.compile(r"[+-]?(\d+(\.\d*)?|\.\d+)([eE][+-]?\d+)?")


def _split_dict_keys(dct: Dict[str, Any]) -> Dict[str, Any]:
    """Splits the given dict keys by first '.' to subdicts."""
    res = {}
    for key, value in dct.items():
        if "." in key:
            outer_key, _, inner_key = key.partition(".")
            if outer_key in res:
                if not isinstance(res[outer_key], dict):
                    raise ValueError(f"Cannot combine {outer_key!r} with {res!r}")
                res[outer_key][inner_key] = value
            else:
                res[outer_key] = {inner_key: value}
        else:
            if key in res:
                raise ValueError(f"Cannot combine {key!r} with {res!r}")
            res[key] = value

    return res


def _isinstance_deep(val: Any, tp_chk: Type) -> bool:
    """Verifies if the given value is an instance of the tp_chk, allowing for typing extensions."""
    if tp_chk is Any:
        return True
    elif typing.get_origin(tp_chk) is Literal:
        (value,) = typing.get_args(val)
        return val == value
    elif typing.get_origin(tp_chk) is list:
        (inner_type,) = typing.get_args(val)
        return isinstance(val, list) and all(_isinstance_deep(v, inner_type) for v in val)
    elif typing.get_origin(tp_chk) is tuple:
        inner_types = typing.get_args(val)
        if len(inner_types) == 2 and inner_types[1] == Ellipsis:
            return isinstance(val, tuple) and all(_isinstance_deep(v, inner_types[0]) for v in val)
        else:
            return (
                isinstance(val, tuple)
                and len(val) == len(inner_types)
                and all(_isinstance_deep(v, inner_type) for v, inner_type in zip(val, inner_types))
            )
    elif typing.get_origin(tp_chk) is dict:
        key_type, value_type = typing.get_args(val)
        return isinstance(val, dict) and all(
            _isinstance_deep(k, key_type) and _isinstance_deep(v, value_type)
            for k, v in val.items()
        )
    else:
        return isinstance(val, tp_chk)



# Copyright (c) 2025, NVIDIA CORPORATION.
# SPDX-License-Identifier: BSD-3-Clause

try:
    from importlib.metadata import version

    __version__ = version("megatron-energon")
except Exception:
    # Fallback for development mode when package is not installed
    __version__ = "unknown"

from megatron.energon.cache import (
    CachePool,
    DecodeFileStore,
    DirectLazy,
    FileCacheLazy,
    FileStore,
    FileStoreCachePool,
    FileStoreDecoder,
    Lazy,
    MockLazy,
    NoCachePool,
    SystemFileStore,
)
from megatron.energon.edataclass import edataclass
from megatron.energon.flavors import (
    BaseCoreDatasetFactory,
    BaseWebdatasetFactory,
    CaptioningSample,
    CaptioningWebdataset,
    CrudeSample,
    CrudeWebdataset,
    DefaultDecoderWebdatasetFactory,
    DefaultGenericWebdatasetFactory,
    ImageClassificationSample,
    ImageClassificationWebdataset,
    ImageSample,
    ImageWebdataset,
    InterleavedSample,
    InterleavedWebdataset,
    JoinedWebdatasetFactory,
    MultiChoiceVQASample,
    MultiChoiceVQAWebdataset,
    OCRSample,
    OCRWebdataset,
    Sample,
    SampleDecoder,
    SavableDataset,
    SimilarityInterleavedSample,
    SimilarityInterleavedWebdataset,
    StandardWebdatasetFactory,
    TextSample,
    TextWebdataset,
    VidQASample,
    VidQAWebdataset,
    VQAOCRWebdataset,
    VQASample,
    VQAWebdataset,
)
from megatron.energon.loader import get_loader, get_savable_loader
from megatron.energon.metadataset import (
    DatasetLoader,
    DatasetLoaderInterface,
    Metadataset,
    MetadatasetV2,
    load_dataset,
    prepare_metadataset,
)
from megatron.energon.savable_loader import SavableDataLoader
from megatron.energon.source_info import SourceInfo
from megatron.energon.task_encoder import (
    AugmentTaskEncoder,
    Batch,
    Cooker,
    DefaultTaskEncoder,
    TaskEncoder,
    basic_sample_keys,
    batch_list,
    batch_pad_stack,
    batch_stack,
    cooker,
    generic_batch,
    get_train_dataset,
    get_val_dataset,
    get_val_datasets,
    stateless,
)
from megatron.energon.worker import WorkerConfig
from megatron.energon.wrappers import (
    BatchDataset,
    BlendDataset,
    ConcatDataset,
    EpochizeDataset,
    FilterDataset,
    GcDataset,
    GroupBatchDataset,
    IterMapDataset,
    LimitDataset,
    LogSampleDataset,
    MapDataset,
    MixBatchDataset,
    PackingDataset,
    RepeatDataset,
    ShuffleBufferDataset,
    SkipSample,
    concat_pad,
    generic_concat,
    homogeneous_concat_mix,
)

__all__ = [
    "__version__",
    "AugmentTaskEncoder",
    "BaseCoreDatasetFactory",
    "BaseWebdatasetFactory",
    "basic_sample_keys",
    "batch_list",
    "batch_pad_stack",
    "batch_stack",
    "Batch",
    "BatchDataset",
    "BlendDataset",
    "CachePool",
    "CaptioningSample",
    "CaptioningWebdataset",
    "concat_pad",
    "ConcatDataset",
    "cooker",
    "Cooker",
    "CrudeSample",
    "CrudeWebdataset",
    "DatasetLoader",
    "DatasetLoaderInterface",
    "DecodeFileStore",
    "DefaultDecoderWebdatasetFactory",
    "DefaultGenericWebdatasetFactory",
    "DefaultTaskEncoder",
    "DirectLazy",
    "EpochizeDataset",
    "edataclass",
    "FileCacheLazy",
    "FileStore",
    "FileStoreCachePool",
    "FileStoreDecoder",
    "FilterDataset",
    "GcDataset",
    "generic_batch",
    "generic_concat",
    "get_loader",
    "get_savable_loader",
    "get_train_dataset",
    "get_val_dataset",
    "get_val_datasets",
    "GroupBatchDataset",
    "homogeneous_concat_mix",
    "ImageClassificationSample",
    "ImageClassificationWebdataset",
    "ImageSample",
    "ImageWebdataset",
    "InterleavedSample",
    "InterleavedWebdataset",
    "IterMapDataset",
    "JoinedWebdatasetFactory",
    "Lazy",
    "LimitDataset",
    "load_dataset",
    "LogSampleDataset",
    "MapDataset",
    "Metadataset",
    "MetadatasetV2",
    "MixBatchDataset",
    "MultiChoiceVQASample",
    "MultiChoiceVQAWebdataset",
    "MockLazy",
    "NoCachePool",
    "OCRSample",
    "OCRWebdataset",
    "PackingDataset",
    "prepare_metadataset",
    "RepeatDataset",
    "Sample",
    "SampleDecoder",
    "SavableDataLoader",
    "SavableDataset",
    "ShuffleBufferDataset",
    "SimilarityInterleavedSample",
    "SimilarityInterleavedWebdataset",
    "SkipSample",
    "SourceInfo",
    "StandardWebdatasetFactory",
    "stateless",
    "SystemFileStore",
    "TaskEncoder",
    "TextSample",
    "TextWebdataset",
    "VidQASample",
    "VidQAWebdataset",
    "VQAOCRWebdataset",
    "VQASample",
    "VQAWebdataset",
    "WorkerConfig",
]



# Copyright (c) 2025, NVIDIA CORPORATION.
# SPDX-License-Identifier: BSD-3-Clause

from dataclasses import dataclass
from typing import Any, Union

from megatron.energon.epathlib import EPath


@dataclass(frozen=True, slots=True, kw_only=True)
class SourceInfo:
    """Information about the source of a sample, i.e. where the data was loaded from."""

    #: The path to the dataset
    dataset_path: EPath
    #: The index of the sample in the dataset
    index: Union[str, int]
    #: The name of the shard tar file
    shard_name: str
    #: The names of the files in the shard used to create the sample
    file_names: tuple[str, ...]


def add_source_info(sample: Any, *source_infos: SourceInfo) -> Any:
    """Adds a source info to a sample."""
    from megatron.energon.flavors.base_dataset import Sample

    if isinstance(sample, Sample) or hasattr(sample, "__sources__"):
        if sample.__sources__ is None:
            sample.__sources__ = source_infos
        else:
            sample.__sources__ = (*sample.__sources__, *source_infos)
    elif isinstance(sample, dict):
        if sample.get("__sources__") is None:
            sample["__sources__"] = source_infos
        else:
            sample["__sources__"] = (*sample["__sources__"], *source_infos)
    return sample



# Copyright (c) 2025, NVIDIA CORPORATION.
# SPDX-License-Identifier: BSD-3-Clause

import os
import random
import sqlite3
import threading
import time
from typing import Any, ClassVar


class ThreadLocalStorage:
    """
    A class that allows to store data in a thread-local storage.

    Example Usage:
    ```python
    class MyThreadLocalStorage(ThreadLocalStorage):
        __thread_local__ = ("my_data",)

        # This is shared across threads
        other_data: int

        # This is local per thread
        my_data: int

        def __thread_init__(self):
            # This is called when the data on a thread is initialized, which has
            # not been accessed yet on that thread to set the value of that data.
            self.my_data = 0

    ```
    """

    __thread_local__: ClassVar[tuple[str, ...]]
    _storage: object

    def __init__(self):
        self._storage = threading.local()

    def __getattribute__(self, name: str) -> Any:
        if name in ("__thread_local__", "_storage"):
            return object.__getattribute__(self, name)
        if name in self.__thread_local__:
            if not self._thread_initialized:
                self._storage.__initialized__ = True
                self.__thread_init__()

            return getattr(self._storage, name)
        return object.__getattribute__(self, name)

    def __delattr__(self, name: str) -> None:
        if name in self.__thread_local__:
            delattr(self._storage, name)
            return
        object.__delattr__(self, name)

    def __setattr__(self, name: str, value: Any) -> None:
        if name in self.__thread_local__:
            if not self._thread_initialized:
                self._storage.__initialized__ = True
                self.__thread_init__()
            setattr(self._storage, name, value)
            return
        object.__setattr__(self, name, value)

    @property
    def _thread_initialized(self) -> bool:
        """Check if the thread has been initialized."""
        return getattr(self._storage, "__initialized__", False)

    def thread_close(self):
        """Close the thread-local storage."""
        if self._thread_initialized:
            delattr(self._storage, "__initialized__")

    def __thread_init__(self):
        """Called when the data on a thread is accessed for the first time, to
        set the initial value of that data."""
        # Copy the data from the default values
        for name in self.__thread_local__:
            try:
                default_value = object.__getattribute__(self, name)
            except AttributeError:
                pass
            else:
                setattr(self._storage, name, default_value)


class ThreadLocalSqlite(ThreadLocalStorage):
    """A class that allows to store data in a thread-local storage."""

    database: str

    __thread_local__ = ("connection", "cursor")

    connection: sqlite3.Connection
    cursor: sqlite3.Cursor

    def __init__(self, database: str):
        super().__init__()
        self.database = database

    def __thread_init__(self):
        """Initialize the connection and cursor."""
        self.connection = sqlite3.connect(self.database)
        self.cursor = self.connection.cursor()
        self.connection.execute("PRAGMA busy_timeout = 5000;")

    def select_one(self, query: str, params: tuple[Any, ...] = ()):
        """Select one row from the database."""
        self.cursor.execute(query, params)
        return self.cursor.fetchone()

    def select_all(self, query: str, params: tuple[Any, ...] = ()):
        """Select all rows from the database."""
        self.cursor.execute(query, params)
        return self.cursor.fetchall()

    def thread_close(self):
        """Close the connection and cursor."""
        if self._thread_initialized:
            self.cursor.close()
            self.connection.close()
            super().thread_close()


def main():
    """Test the ThreadLocalSqlite class."""
    import concurrent.futures

    sqlite = ThreadLocalSqlite("tmp.sqlite")
    sqlite.cursor.execute("CREATE TABLE IF NOT EXISTS test (id INTEGER PRIMARY KEY, name TEXT)")
    sqlite.cursor.execute("INSERT INTO test (name) VALUES (?)", ("test",))
    sqlite.connection.commit()

    def _test_thread_local(sqlite_thread_local: ThreadLocalSqlite):
        time.sleep(random.random())
        print(sqlite_thread_local.select_all("SELECT * FROM test"))

    with concurrent.futures.ThreadPoolExecutor(max_workers=10) as executor:
        futures = []
        for _ in range(20):
            futures.append(executor.submit(_test_thread_local, sqlite))
        for future in concurrent.futures.as_completed(futures):
            future.result()

    os.remove("tmp.sqlite")


if __name__ == "__main__":
    main()



# Copyright (c) 2025, NVIDIA CORPORATION.
# SPDX-License-Identifier: BSD-3-Clause

import inspect
from typing import (
    Any,
    Callable,
    Dict,
    Generator,
    Generic,
    Iterator,
    List,
    Optional,
    Tuple,
    TypeVar,
    Union,
)

from megatron.energon.errors import SYSTEM_EXCEPTIONS, FatalSampleError
from megatron.energon.flavors.base_dataset import SavableDataset, set_sample_restore_key
from megatron.energon.source_info import SourceInfo
from megatron.energon.worker import WorkerConfig
from megatron.energon.wrappers._log_exception import log_exception
from megatron.energon.wrappers.base import BaseWrapperDataset, SampleIndex, get_sample_restore_key
from megatron.energon.wrappers.skip import SkipSample

T_batch = TypeVar("T_batch", covariant=True)
T_batch_sample = TypeVar("T_batch_sample", covariant=True)


class BatchDataset(BaseWrapperDataset[T_batch_sample, T_batch], Generic[T_batch_sample, T_batch]):
    """This dataset wrapper transforms a dataset of samples into a dataset of batches."""

    batch_size: int
    batcher: Callable[[List[T_batch_sample]], T_batch]
    drop_last: bool
    error_handler: Callable[[Exception, list[T_batch_sample], list[SourceInfo]], None]
    _sample_index: SampleIndex
    _generator_sample_keys: Optional[Any]
    _generator_offset: Optional[int]

    _savable_fields = ("_sample_index", "_generator_sample_keys", "_generator_offset")

    def __init__(
        self,
        dataset: SavableDataset[T_batch_sample],
        batch_size: int,
        batcher: Callable[[List[T_batch_sample]], T_batch],
        *,
        batcher_stateless: bool = False,
        batcher_config: Optional[Union[Dict[str, Any], Callable[[], Dict[str, Any]]]] = None,
        drop_last: bool = False,
        error_handler: Callable[
            [Exception, List[T_batch_sample], List[SourceInfo]], None
        ] = log_exception,
        failure_tolerance: Optional[int] = 100,
        worker_config: WorkerConfig,
    ):
        """Construct a BatchDataset.

        Args:
            dataset: The input dataset to wrap
            batch_size: The desired batch size. The last batch may be smaller.
            batcher: Function which combines separate samples into a single object. May raise
                :exc:`megatron.energon.SkipSample` to skip a sample.
            batcher_stateless: If True, the batcher is stateless, thus samples can be stored/
                restored.
            batcher_config: Configuration for the batcher function. If callable, it should return the
                configuration. Defaults to None.
            drop_last: If True, the last batch is dropped if it is smaller than the batch size.
            error_handler: Function which handles exceptions raised by the batcher. The default
                implementation logs the exception.
            failure_tolerance: The number of consecutive failures after which the dataset is considered broken.
            worker_config: Configuration for the workers.
        """
        super().__init__(dataset, worker_config=worker_config)
        self.batch_size = batch_size
        self.batcher = batcher
        self.batcher_stateless = batcher_stateless
        self.batcher_config = batcher_config
        self.drop_last = drop_last
        self.error_handler = error_handler
        self.failure_tolerance = failure_tolerance

        self.reset_state_own()

    def reset_state_own(self) -> None:
        self._sample_index = SampleIndex(self.worker_config, src=self)
        self._generator_sample_keys = None
        self._generator_offset = None

    def __len__(self):
        n_samples = len(self.dataset)
        num_workers = max(self.worker_config.num_workers, 1)
        n_samples_per_worker_floor = n_samples // num_workers
        remaining_n_sample_workers = n_samples % num_workers
        n_batches_per_worker_floor = n_samples_per_worker_floor // self.batch_size
        if n_samples_per_worker_floor % self.batch_size != 0 and not self.drop_last:
            n_batches_per_worker_floor += 1
        # Correct number of batches for the workers which yield 1 more sample (to balance)
        n_batches_per_worker_ceil = (n_samples_per_worker_floor + 1) // self.batch_size
        if n_batches_per_worker_ceil % self.batch_size != 0 and not self.drop_last:
            n_batches_per_worker_ceil += 1

        return (
            n_batches_per_worker_floor * (num_workers - remaining_n_sample_workers)
            + n_batches_per_worker_ceil * remaining_n_sample_workers
        )

    def __iter__(self) -> Iterator[T_batch]:
        batch: List[T_batch_sample] = []
        sample_restore_keys = []

        last_batch_failures = 0

        if self._generator_sample_keys is not None:
            sample_restore_keys = self._generator_sample_keys
            assert self._generator_offset is not None
            batch = [self.dataset.restore_sample(inner_idx) for inner_idx in sample_restore_keys]
            with self._sample_index.ctx(self._sample_index.current_idx) as sample_idx:
                batch_sample = self.batcher(batch)
            assert isinstance(batch_sample, Generator)
            assert inspect.isgeneratorfunction(self.batcher), (
                f"Generator in {self.batcher} but not marked as such."
            )
            target_offset = self._generator_offset
            self._generator_offset = 0
            for batch_sub_idx, (sample_idx, inner_batch_sample) in enumerate(
                self._sample_index.iter_ctx(batch_sample, sample_idx)
            ):
                # Skip other samples
                if batch_sub_idx >= target_offset:
                    self._generator_offset = batch_sub_idx + 1
                    yield set_sample_restore_key(
                        inner_batch_sample,
                        sample_idx,
                        batch_sub_idx,
                        *sample_restore_keys,
                        src=self,
                    )
            self._generator_sample_keys = None
            self._generator_offset = None
            batch.clear()
            sample_restore_keys = []

        def flush():
            nonlocal last_batch_failures

            try:
                with self._sample_index.ctx() as sample_idx:
                    batch_sample = self.batcher(batch)
                if isinstance(batch_sample, Generator):
                    assert inspect.isgeneratorfunction(self.batcher), (
                        f"Generator in {self.batcher} but not marked as such."
                    )
                    self._generator_sample_keys = sample_restore_keys
                    self._generator_offset = 0
                    for batch_sub_idx, (sample_idx, inner_batch_sample) in enumerate(
                        self._sample_index.iter_ctx(batch_sample, sample_idx)
                    ):
                        last_batch_failures = 0
                        self._generator_offset = batch_sub_idx + 1
                        yield set_sample_restore_key(
                            inner_batch_sample,
                            sample_idx,
                            batch_sub_idx,
                            *sample_restore_keys,
                            src=self,
                        )
                    self._generator_sample_keys = None
                    self._generator_offset = None
                else:
                    last_batch_failures = 0
                    set_sample_restore_key(batch_sample, sample_idx, *sample_restore_keys, src=self)
                    yield batch_sample
            except GeneratorExit:
                raise
            except SkipSample:
                pass
            except SYSTEM_EXCEPTIONS:
                raise FatalSampleError.from_sample(batch)
            except Exception as e:
                self.error_handler(e, batch)
                last_batch_failures += 1
                if (
                    self.failure_tolerance is not None
                    and last_batch_failures >= self.failure_tolerance
                ):
                    raise FatalSampleError.from_sample(
                        batch,
                        f"BatchDataset {self.batcher} failed {last_batch_failures} times in a row. Likely your code or dataset are broken.",
                    )
            finally:
                sample_restore_keys.clear()

        for sample in self.dataset:
            batch.append(sample)
            sample_restore_keys.append(get_sample_restore_key(sample))
            if len(batch) == self.batch_size:
                yield from flush()
                batch = []
        if len(batch) > 0 and not self.drop_last:
            yield from flush()

    def can_restore_sample(self) -> bool:
        # Cannot really verify if the returned elements contain a __restore_key__.
        # If the user wants to use this, well...
        return super().can_restore_sample() and self.batcher_stateless

    def assert_can_restore(self) -> None:
        assert self.batcher_stateless, (
            f"Batcher {self.batcher} must be stateless to restore samples"
        )
        super().assert_can_restore()

    def restore_sample(self, restore_key: Tuple[Union[str, int, tuple], ...]) -> T_batch:
        # We need to store multiple indices to restore a batch.
        self.assert_can_restore()
        if inspect.isgeneratorfunction(self.batcher):
            id, sample_idx, batch_sub_idx, *samples_restore_keys = restore_key
            assert id == type(self).__name__
        else:
            id, sample_idx, *samples_restore_keys = restore_key
            assert id == type(self).__name__
        batch = [self.dataset.restore_sample(inner_idx) for inner_idx in samples_restore_keys]
        with self._sample_index.ctx(sample_idx):
            batch_sample = self.batcher(batch)
        if isinstance(batch_sample, Generator):
            assert inspect.isgeneratorfunction(self.batcher), (
                f"Generator in {self.batcher} but not marked as such."
            )
            for cur_batch_sub_idx, (sample_idx, inner_batch_sample) in enumerate(
                self._sample_index.iter_ctx(batch_sample, sample_idx)
            ):
                if cur_batch_sub_idx == batch_sub_idx:
                    return set_sample_restore_key(
                        inner_batch_sample,
                        sample_idx,
                        batch_sub_idx,
                        *samples_restore_keys,
                        src=self,
                    )
            assert False, f"Batch sub-index {batch_sub_idx} not found in batch"
        else:
            return set_sample_restore_key(
                batch_sample,
                sample_idx,
                *samples_restore_keys,
                src=self,
            )

    def config(self) -> Dict[str, Any]:
        return {
            "type": type(self).__qualname__,
            "batch_size": self.batch_size,
            "batcher": self._function_config(self.batcher),
            **(
                {
                    "batcher_config": (
                        self.batcher_config()
                        if callable(self.batcher_config)
                        else self.batcher_config
                    )
                }
                if self.batcher_config
                else {}
            ),
            "batcher_stateless": self.batcher_stateless,
            "drop_last": self.drop_last,
            "error_handler": self._function_config(self.error_handler),
            "worker_config": self.worker_config.config(),
            "dataset": self.dataset.config(),
        }

    def __str__(self):
        return f"BatchDataset(batch_size={self.batch_size}, drop_last={self.drop_last}, batcher={self.batcher}, dataset={self.dataset})"



# Copyright (c) 2025, NVIDIA CORPORATION.
# SPDX-License-Identifier: BSD-3-Clause

import math
from typing import Any, Dict, Generic, Iterator, Optional, TypeVar, Union

from megatron.energon.flavors.base_dataset import SavableDataset
from megatron.energon.worker import WorkerConfig
from megatron.energon.wrappers.base import BaseWrapperDataset

T_sample = TypeVar("T_sample")


class RepeatDataset(BaseWrapperDataset[T_sample, T_sample], Generic[T_sample]):
    """This dataset repeats the inner dataset indefinitely or a specific number of repeats."""

    repeats: Optional[Union[int, float]]
    _repetition: int
    _index: int

    _savable_fields = ("_repetition", "_index")

    def __init__(
        self,
        dataset: SavableDataset[T_sample],
        *,
        repeats: Optional[Union[int, float]] = None,
        restart: bool = True,
        worker_config: WorkerConfig,
    ):
        """Construct a RepeatDataset.

        Args:
            dataset: The input dataset to repeat.
            repeats: Number of repeats, `None` for indefinitely repeating.
            restart: If true, restart the underlying dataset after iterating once through the
                repeats if repeats is set to an integer, but still stop iterating.
            worker_config: Configuration for the workers.
        """
        super().__init__(dataset, worker_config=worker_config)
        self.repeats = repeats
        self.restart = restart

        self.reset_state_own()

    def reset_state_own(self) -> None:
        self._repetition = 0
        self._index = 0

    def __len__(self):
        if self.repeats is None:
            return len(self.dataset)
        return int(len(self.dataset) * self.repeats)

    def __iter__(self) -> Iterator[T_sample]:
        assert self.repeats is not None or self.dataset.worker_has_samples(), (
            "Cannot repeat empty dataset indefinitely"
        )

        ds_len = len(self.dataset)

        while self.repeats is None or self._repetition < self.repeats:
            if self.repeats is not None and self._repetition == math.floor(self.repeats):
                # Last iteration, adjust the number of samples
                fraction = self.repeats - math.floor(self.repeats)
                stop_after = math.floor(ds_len * fraction)
                if self._index >= stop_after:
                    # We restored an index and it is already past the stop_after
                    break
            else:
                stop_after = None

            for sample in self.dataset:
                self._index += 1
                yield sample
                if stop_after is not None and self._index >= stop_after:
                    break

            if self.worker_config.should_log(level=2):
                self.worker_config.worker_log(
                    {
                        "t": "RepeatDataset.repeat",
                        "r": self.worker_config.rank,
                        "w": self.worker_config.rank_worker_id(),
                        "offset": self._repetition,
                        "repeats": self.repeats,
                    }
                )
            self._repetition += 1
            self._index = 0

        if self.restart:
            self._repetition = 0
        else:
            # No more repeats
            self._repetition = math.ceil(self.repeats)

    def config(self) -> Dict[str, Any]:
        return {
            "type": type(self).__qualname__,
            "dataset": self.dataset.config(),
            "repeats": self.repeats,
            "worker_config": self.worker_config.config(),
        }

    def __str__(self):
        return f"RepeatDataset(repeats={self.repeats}, dataset={self.dataset})"
